# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Session class: A session to the HMC, optionally in context of an HMC user.
"""


import json
import time
import re
from copy import copy
from collections import OrderedDict
from collections.abc import Iterable
import requests
import urllib3

from ._exceptions import HTTPError, ServerAuthError, ClientAuthError, \
    ParseError, ConnectTimeout, ReadTimeout, RetriesExceeded, \
    OperationTimeout, Error
from ._exceptions import ConnectionError  # pylint: disable=redefined-builtin

from ._timestats import TimeStatsKeeper
from ._auto_updater import AutoUpdater
from ._logging import get_logger, logged_api_call
from ._constants import DEFAULT_CONNECT_TIMEOUT, DEFAULT_CONNECT_RETRIES, \
    DEFAULT_READ_TIMEOUT, DEFAULT_READ_RETRIES, DEFAULT_MAX_REDIRECTS, \
    DEFAULT_OPERATION_TIMEOUT, DEFAULT_STATUS_TIMEOUT, \
    DEFAULT_NAME_URI_CACHE_TIMETOLIVE, HMC_LOGGER_NAME, \
    HTML_REASON_WEB_SERVICES_DISABLED, HTML_REASON_OTHER, \
    DEFAULT_HMC_PORT, BLANKED_OUT_STRING
from ._utils import repr_obj_id
from ._version import __version__

__all__ = ['Session', 'Job', 'RetryTimeoutConfig', 'get_password_interface']

HMC_LOGGER = get_logger(HMC_LOGGER_NAME)

_HMC_SCHEME = "https"
_STD_HEADERS = {
    'User-Agent': f'python-zhmcclient/{__version__}',
    'Content-type': 'application/json',
    'Accept': '*/*'
}

# Properties whose values are always blanked out in the HMC log entries
BLANKED_OUT_PROPERTIES = [
    'boot-ftp-password',    # partition create/update
    'bind-password',        # LDAP server def. create/update
    'ssc-master-pw',        # image profile cr/upd, part. cr/upd, LPAR upd
    'password',             # user create/update
    'zaware-master-pw',     # image profile create/update, LPAR update
]


def _handle_request_exc(exc, retry_timeout_config):
    """
    Handle a :exc:`request.exceptions.RequestException` exception that was
    raised.
    """
    if isinstance(exc, requests.exceptions.ConnectTimeout):
        new_exc = ConnectTimeout(_request_exc_message(exc), exc,
                                 retry_timeout_config.connect_timeout,
                                 retry_timeout_config.connect_retries)
        new_exc.__cause__ = None
        raise new_exc  # ConnectTimeout

    if isinstance(exc, requests.exceptions.ReadTimeout):
        new_exc = ReadTimeout(_request_exc_message(exc), exc,
                              retry_timeout_config.read_timeout,
                              retry_timeout_config.read_retries)
        new_exc.__cause__ = None
        raise new_exc  # ReadTimeout

    if isinstance(exc, requests.exceptions.RetryError):
        new_exc = RetriesExceeded(_request_exc_message(exc), exc,
                                  retry_timeout_config.connect_retries)
        new_exc.__cause__ = None
        raise new_exc  # RetriesExceeded

    new_exc = ConnectionError(_request_exc_message(exc), exc)
    new_exc.__cause__ = None
    raise new_exc  # ConnectionError


def _request_exc_message(exc):
    """
    Return a reasonable exception message from a
    :exc:`request.exceptions.RequestException` exception.

    The approach is to dig deep to the original reason, if the original
    exception is present, skipping irrelevant exceptions such as
    `urllib3.exceptions.MaxRetryError`, and eliminating useless object
    representations such as the connection pool object in
    `urllib3.exceptions.NewConnectionError`.

    Parameters:

      exc (:exc:`~request.exceptions.RequestException`): Exception

    Returns:

      string: A reasonable exception message from the specified exception.
    """
    messages = []
    for arg in exc.args:

        if isinstance(arg, Exception):
            org_exc = arg
            if isinstance(org_exc, urllib3.exceptions.MaxRetryError):
                message = f"{org_exc}, reason: {org_exc.reason}"
            else:
                message = str(org_exc)
        else:
            message = str(arg)

        # Eliminate useless object repr at begin of the message
        m = re.match(r'^(\(<[^>]+>, \'(.*)\'\)|<[^>]+>: (.*))$', message)
        if m:
            message = m.group(2) or m.group(3)

        messages.append(message)

    return ", ".join(messages)


class RetryTimeoutConfig:
    # pylint: disable=too-few-public-methods
    """
    A configuration setting that specifies verious retry counts and timeout
    durations.

    HMC/SE version requirements: None
    """

    def __init__(self, connect_timeout=None, connect_retries=None,
                 read_timeout=None, read_retries=None, max_redirects=None,
                 operation_timeout=None, status_timeout=None,
                 name_uri_cache_timetolive=None):
        """
        For all parameters, `None` means that this object does not specify a
        value for the parameter, and that a default value should be used
        (see :ref:`Constants`).

        All parameters are available as instance attributes.

        Parameters:

          connect_timeout (:term:`number`): Connect timeout in seconds.
            This timeout applies to making a connection at the socket level.
            The same socket connection is used for sending an HTTP request to
            the HMC and for receiving its HTTP response.
            The special value 0 means that no timeout is set.

          connect_retries (:term:`integer`): Number of retries (after the
            initial attempt) for connection-related issues. These retries are
            performed for failed DNS lookups, failed socket connections, and
            socket connection timeouts.

          read_timeout (:term:`number`): Read timeout in seconds.
            This timeout applies to reading at the socket level, when receiving
            an HTTP response.
            The special value 0 means that no timeout is set.

          read_retries (:term:`integer`): Number of retries (after the
            initial attempt) for read-related issues. These retries are
            performed for failed socket reads and socket read timeouts.
            A retry consists of resending the original HTTP request. The
            zhmcclient restricts these retries to just the HTTP GET method.
            For other HTTP methods, no retry will be performed.

          max_redirects (:term:`integer`): Maximum number of HTTP redirects.

          operation_timeout (:term:`number`): Asynchronous operation timeout in
            seconds. This timeout applies when waiting for the completion of
            asynchronous HMC operations. The special value 0 means that no
            timeout is set.

          status_timeout (:term:`number`): Resource status timeout in seconds.
            This timeout applies when waiting for the transition of the status
            of a resource to a desired status. The special value 0 means that
            no timeout is set.

          name_uri_cache_timetolive (:term:`number`): Time to the next
            automatic invalidation of the Name-URI cache of manager objects, in
            seconds since the last invalidation. The special value 0 means
            that no Name-URI cache is maintained (i.e. the caching is
            disabled).
        """
        self.connect_timeout = connect_timeout
        self.connect_retries = connect_retries
        self.read_timeout = read_timeout
        self.read_retries = read_retries
        self.max_redirects = max_redirects
        self.operation_timeout = operation_timeout
        self.status_timeout = status_timeout
        self.name_uri_cache_timetolive = name_uri_cache_timetolive

        # Read retries only for these HTTP methods:
        self.allowed_methods = {'GET'}

    _attrs = ('connect_timeout', 'connect_retries', 'read_timeout',
              'read_retries', 'max_redirects', 'operation_timeout',
              'status_timeout', 'name_uri_cache_timetolive',
              'allowed_methods')

    def override_with(self, override_config):
        """
        Return a new configuration object that represents the configuration
        from this configuration object acting as a default, and the specified
        configuration object overriding that default for any of its
        attributes that are not `None`.

        Parameters:

          override_config (:class:`~zhmcclient.RetryTimeoutConfig`):
            The configuration object overriding the defaults defined in this
            configuration object.

        Returns:

          :class:`~zhmcclient.RetryTimeoutConfig`:
            A new configuration object representing this configuration object,
            overridden by the specified configuration object.
        """
        ret = RetryTimeoutConfig()
        for attr in RetryTimeoutConfig._attrs:
            value = getattr(self, attr)
            if override_config and getattr(override_config, attr) is not None:
                value = getattr(override_config, attr)
            setattr(ret, attr, value)
        return ret


def get_password_interface(host, userid):
    """
    Interface to the password retrieval function that is invoked by
    :class:`~zhmcclient.Session` if no password is provided.

    Parameters:

      host (string): Hostname or IP address of the HMC
      userid (string): Userid on the HMC

    Returns:

      string: Password of the userid on the HMC
    """
    raise NotImplementedError


def _headers_for_logging(headers):
    """
    Return the input headers dict with blanked out values for any headers that
    carry sensitive information, so that it can be logged or displayed.

    The headers argument is not modified; if it needs to be changed, a copy is
    made that is changed.
    """
    if headers and 'X-API-Session' in headers:
        headers = headers.copy()
        headers['X-API-Session'] = BLANKED_OUT_STRING
    return headers


class Session:
    """
    A session to the HMC, optionally in context of an HMC user.

    The session supports operations that require to be authenticated, as well
    as operations that don't (e.g. obtaining the API version).

    The session can keep statistics about the elapsed time for issuing HTTP
    requests against the HMC API. Instance variable
    :attr:`~zhmcclient.Session.time_stats_keeper` is used to enable/disable the
    measurements, and to print the statistics.

    HMC/SE version requirements: None
    """

    default_rt_config = RetryTimeoutConfig(
        connect_timeout=DEFAULT_CONNECT_TIMEOUT,
        connect_retries=DEFAULT_CONNECT_RETRIES,
        read_timeout=DEFAULT_READ_TIMEOUT,
        read_retries=DEFAULT_READ_RETRIES,
        max_redirects=DEFAULT_MAX_REDIRECTS,
        operation_timeout=DEFAULT_OPERATION_TIMEOUT,
        status_timeout=DEFAULT_STATUS_TIMEOUT,
        name_uri_cache_timetolive=DEFAULT_NAME_URI_CACHE_TIMETOLIVE,
    )

    def __init__(self, host, userid=None, password=None, session_id=None,
                 get_password=None, retry_timeout_config=None,
                 port=DEFAULT_HMC_PORT, verify_cert=True):
        # pylint: disable=line-too-long
        """
        Creating a session object will not immediately cause a logon to be
        attempted; the logon is deferred until needed.

        There are several alternatives for specifying the authentication
        related parameters:

        * `userid`/`password` only: The session is initially in a logged-off
          state and subsequent operations that require logon will use the
          specified userid and password to automatically log on. The returned
          session-id will be stored in this session object. Subsequent
          operations that require logon will use that session-id. Once the HMC
          expires that session-id, subsequent operations that require logon
          will cause a re-logon with the specified userid and password.

        * `userid`/`password` and `session_id`: The specified session-id will
          be stored in this session object, so that the session is initially in
          a logged-on state. Subsequent operations that require logon will use
          that session-id. Once the HMC expires that session-id, subsequent
          operations that require logon will cause a re-logon with the
          specified userid/password.
          In this case, the `host` parameter must specify the single HMC that
          has that session.

        * `session_id` only: The specified session-id will be stored in this
          session object, so that the session is initially in a logged-on
          state. Subsequent operations that require logon will use the stored
          session-id. Once the HMC expires the session-id, subsequent
          operations that require logon will cause an
          :exc:`~zhmcclient.ServerAuthError` to be raised (because
          userid/password have not been specified, so an automatic re-logon is
          not possible).
          In this case, the `host` parameter must specify the single HMC that
          has that session.

        * Neither `userid`/`password` nor `session_id`: Only operations that do
          not require logon, are possible.

        Parameters:

          host (:term:`string` or iterable of :term:`string`):
            HMC host or list of HMC hosts to try from.
            For valid formats, see the :attr:`~zhmcclient.Session.host`
            property.
            If `session_id` is specified, this must be the single HMC that has
            that session.
            Must not be `None`.

          userid (:term:`string`):
            Userid of the HMC user to be used, or `None`.

          password (:term:`string`):
            Password of the HMC user to be used, if `userid` was specified.

          session_id (:term:`string`):
            Session-id to be used for this session, or `None`.

          get_password (:term:`callable`):
            A password retrieval function, or `None`.

            If provided, this function will be called if a password is needed
            but not provided. This mechanism can be used for example by command
            line interfaces for prompting for the password.

            The password retrieval function must follow the interface
            defined in :func:`~zhmcclient.get_password_interface`.

          retry_timeout_config (:class:`~zhmcclient.RetryTimeoutConfig`):
            The retry/timeout configuration for this session for use by any of
            its HMC operations, overriding any defaults.

            `None` for an attribute in that configuration object means that the
            default value will be used for that attribute.

            `None` for the entire `retry_timeout_config` parameter means that a
            default configuration will be used with the default values for all
            of its attributes.

            See :ref:`Constants` for the default values.

          port (:term:`integer`):
            HMC TCP port. Defaults to
            :attr:`~zhmcclient._constants.DEFAULT_HMC_PORT`.
            For details, see the :attr:`~zhmcclient.Session.port` property.

          verify_cert (bool or :term:`string`):
            Controls whether and how the client verifies the server certificate
            presented by the HMC during SSL/TLS handshake:

            * `False`: Do not verify the HMC certificate. Not verifying the HMC
              certificate means the zhmcclient will not detect hostname
              mismatches, expired certificates, revoked certificates, or
              otherwise invalid certificates. Since this mode makes the
              connection vulnerable to man-in-the-middle attacks, it is insecure
              and should not be used in production environments.

            * `True`: Verify the HMC certificate using the CA certificates from
              the first of these locations:

              - The file or directory in the REQUESTS_CA_BUNDLE env.var, if set
              - The file or directory in the CURL_CA_BUNDLE env.var, if set
              - The Python 'certifi' package (which contains the
                `Mozilla Included CA Certificate List <https://wiki.mozilla.org/CA/Included_Certificates>`_).

            * :term:`string`: Path name of a certificate file or directory.
              Verify the HMC certificate using the CA certificates in that file
              or directory.

            For details, see the :ref:`HMC certificate` section.

            *Added in version 0.31*
        """  # noqa: E501
        # pylint: enable=line-too-long

        if isinstance(host, str):
            self._hosts = [host]
        else:
            self._hosts = list(host)
            assert len(self._hosts) >= 1
        self._port = port
        self._userid = userid
        self._password = password
        self._verify_cert = verify_cert
        self._get_password = get_password
        self._retry_timeout_config = self.default_rt_config.override_with(
            retry_timeout_config)
        self._headers = copy(_STD_HEADERS)  # dict with standard HTTP headers
        if session_id is not None:
            # Create a logged-on state (nearly same state as in _do_logon())
            self._session_id = session_id
            self._session = self._new_session(self.retry_timeout_config)
            self._headers['X-API-Session'] = session_id
            assert len(self._hosts) == 1
            self._actual_host = self._hosts[0]
            self._base_url = \
                self._create_base_url(self._actual_host, self._port)
            # The following are set in _do_logon()) but not here:
            self._session_credential = None
            self._object_topic = None
            self._job_topic = None
        else:
            # Create a logged-off state (same state as in _do_logoff())
            self._session_id = None
            self._session = None
            self._actual_host = None
            self._base_url = None
            self._session_credential = None
            self._object_topic = None
            self._job_topic = None
        self._time_stats_keeper = TimeStatsKeeper()
        self._auto_updater = AutoUpdater(self)

    def __repr__(self):
        """
        Return a string with the state of this session, for debug purposes.
        """
        headers = _headers_for_logging(self.headers)
        ret = (
            f"{repr_obj_id(self)} (\n"
            f"  _hosts={self._hosts!r},\n"
            f"  _userid={self._userid!r},\n"
            f"  _password='...',\n"
            f"  _verify_cert={self._verify_cert!r},\n"
            f"  _get_password={self._get_password!r},\n"
            f"  _retry_timeout_config={self._retry_timeout_config!r},\n"
            f"  _actual_host={self._actual_host!r},\n"
            f"  _base_url={self._base_url!r},\n"
            f"  _headers={headers!r},\n"
            f"  _session_id={BLANKED_OUT_STRING!r},\n"
            f"  _session={self._session!r}\n"
            f"  _object_topic={self._object_topic!r}\n"
            f"  _job_topic={self._job_topic!r}\n"
            f"  _auto_updater={self._auto_updater!r}\n"
            ")")
        return ret

    @property
    def host(self):
        """
        :term:`string` or list of :term:`string`: HMC host or redundant HMC
        hosts to use. The first working HMC from this list will actually be
        used. The working state of the HMC is detrmined using the
        'Query API Version' operation for which no authentication is needed.

        Each host will be in one of the following formats:

          * a short or fully qualified DNS hostname
          * a literal (= dotted) IPv4 address
          * a literal IPv6 address, formatted as defined in :term:`RFC3986`
            with the extensions for zone identifiers as defined in
            :term:`RFC6874`, supporting ``-`` (minus) for the delimiter
            before the zone ID string, as an additional choice to ``%25``
        """
        if len(self._hosts) == 1:
            host = self._hosts[0]
        else:
            host = self._hosts
        return host

    @property
    def actual_host(self):
        """
        :term:`string` or `None`: The HMC host that is actually used for this
        session, if the session is in the logged-on state. `None`, if the
        session is in the logged-off state.

        The HMC host will be in one of the following formats:

          * a short or fully qualified DNS hostname
          * a literal (= dotted) IPv4 address
          * a literal IPv6 address, formatted as defined in :term:`RFC3986`
            with the extensions for zone identifiers as defined in
            :term:`RFC6874`, supporting ``-`` (minus) for the delimiter
            before the zone ID string, as an additional choice to ``%25``
        """
        return self._actual_host

    @property
    def port(self):
        """
        :term:`integer`: HMC TCP port that is used for this session.
        """
        return self._port

    @property
    def userid(self):
        """
        :term:`string`: HMC userid that is used for this session.

        If `None`, only operations that do not require authentication, can be
        performed.
        """
        return self._userid

    @property
    def verify_cert(self):
        """
        bool or :term:`string`: Controls whether and how the client verifies
        server certificate presented by the HMC during SSL/TLS handshake.

        For details, see the same-named init parameter.
        """
        return self._verify_cert

    @property
    def get_password(self):
        """
        The password retrieval function, or `None`.

        The password retrieval function must follow the interface defined in
        :func:`~zhmcclient.get_password_interface`.
        """
        return self._get_password

    @property
    def retry_timeout_config(self):
        """
        :class:`~zhmcclient.RetryTimeoutConfig`: The effective retry/timeout
        configuration for this session for use by any of its HMC operations,
        taking into account the defaults and the session-specific overrides.
        """
        return self._retry_timeout_config

    @property
    def base_url(self):
        """
        :term:`string` or `None`: Base URL of the HMC that is actually used for
        this session, if the session is in the logged-on state. `None`, if the
        session is in the logged-off state.

        Example:

        .. code-block:: text

            https://myhmc.acme.com:6794
        """
        return self._base_url

    @property
    def headers(self):
        """
        :term:`header dict`: HTTP headers that are used in requests sent
        for this session.

        Initially, this is the following set of headers:

        .. code-block:: text

            Content-type: application/json
            Accept: */*

        When the session is logged on to the HMC, the session token is added
        to these headers:

        .. code-block:: text

            X-API-Session: ...
        """
        return self._headers

    @property
    def time_stats_keeper(self):
        """
        The time statistics keeper (for a usage example, see section
        :ref:`Time Statistics`).
        """
        return self._time_stats_keeper

    @property
    def session_id(self):
        """
        :term:`string` or `None`: Session ID (= HMC session token) used for this
        session, if the session is in the logged-on state. `None`, if the
        session is in the logged-off state.

        In the logged-off state, any request that requires logon will first
        cause a session to be created on the HMC and will store the session ID
        returned by the HMC in this property.

        In the logged-on state, the session ID stored in this property will be
        used for any requests to the HMC.
        """
        return self._session_id

    @property
    def session_credential(self):
        """
        :term:`string` or `None`: Session credential for this session returned
        by the HMC, if the session is in the logged-on state. `None`, if the
        session is in the logged-off state.
        """
        return self._session_credential

    @property
    def session(self):
        """
        :term:`string` or `None`: :class:`requests.Session` object for this
        session, if the session is in the logged-on state. `None`, if the
        session is in the logged-off state.
        """
        return self._session

    @property
    def object_topic(self):
        """
        :term:`string` or `None`: Name of the notification topic the HMC will
        use to send object-related notification messages to this session, if
        the session is in the logged-on state. `None`, if the session is in the
        logged-off state.

        The associated topic type is "object-notification".
        """
        return self._object_topic

    @property
    def job_topic(self):
        """
        :term:`string` or `None`: Name of the notification topic the HMC will
        use to send job notification messages to this session, if the session
        is in the logged-on state. `None`, if the session is in the logged-off
        state.

        The associated topic type is "job-notification".
        """
        return self._job_topic

    @property
    def auto_updater(self):
        """
        :class:`~zhmcclient.AutoUpdater`: Updater for
        :ref:`auto-updating` of resource and manager objects.
        """
        return self._auto_updater

    @logged_api_call
    def logon(self, verify=False):
        """
        Make sure this session object is logged on to the HMC.

        If `verify=False`, this method determines the logged-on state of this
        session object based on whether there is a session ID set in this
        session object. If a session ID is set, it is assumed to be valid and
        no new session is created on the HMC. Otherwise, a new session will be
        created on the HMC.

        If `verify=True`, this method determines the logged-on state of this
        session object in addition by performing a read operation on the HMC
        that requires to be logged on but no specific authorizations. If a
        session ID is set and if that operation succeeds, no new session is
        created on the HMC. Any failure of that read operation will be ignored.
        Otherwise, a new session will be created on the HMC.

        When a new session has been successfully created on the HMC, the
        :attr:`session_id` attribute of this session object will be set to the
        session ID returned by the HMC to put it into the logged-on state.

        Any exceptions raised from this method are always related to the
        creation of a new session on the HMC - any failures of the read
        operation in the verification case are always ignored.

        Parameters:

          verify (bool): Verify the validity of an existing session ID.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.ClientAuthError`
          :exc:`~zhmcclient.ServerAuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        need_logon = False
        if self._session_id is None:
            need_logon = True
        elif verify:
            try:
                self.get('/api/console', logon_required=False,
                         renew_session=False)
            except Error:
                need_logon = True
        if need_logon:
            self._do_logon()

    @logged_api_call
    def logoff(self, verify=False):
        # pylint: disable=unused-argument
        """
        Make sure this session object is logged off from the HMC.

        If a session ID is set in this session object, its session will be
        deleted on the HMC. If that delete operation fails due to an invalid
        session ID, that failure will be ignored. Any other failures of that
        delete operation will be raised as exceptions.

        When the session has been successfully deleted on the HMC, the
        :attr:`session_id` attribute of this session object will be set to
        `None` to put it into the logged-off state.

        Parameters:

          verify (bool): Deprecated: This parameter will be ignored.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.ConnectionError`
        """
        if self._session_id:
            self._do_logoff()

    @logged_api_call
    def is_logon(self, verify=False):
        """
        Return a boolean indicating whether this session object is logged on to
        the HMC.

        If `verify=False`, this method determines the logged-on state based
        on whether there is a session ID set in this object. If a session ID is
        set, it is assumed to be valid, and `True` is returned. Otherwise,
        `False` is returned. In that case, no exception is ever raised.

        If `verify=True`, this method determines the logged-on state in
        addition by verifying a session ID that is set, by performing a read
        operation on the HMC that requires to be logged on but no specific
        authorizations. If a session ID is set and if that read operation
        succeeds, `True` is returned. If no session ID is set or if that read
        operation fails due to an invalid session ID, `False` is returned.
        Any other failures of that read operation are raised as exceptions,
        because that indicates that a verification with the HMC could not be
        performed.

        Parameters:

          verify (bool): Verify the validity of an existing session ID.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.ConnectionError`
        """
        if self._session_id is None:
            return False
        if verify:
            try:
                self.get('/api/console', logon_required=False,
                         renew_session=False)
            except ServerAuthError:
                return False
        return True

    def _do_logon(self):
        """
        Log on, unconditionally. This can be used to re-logon.
        This requires credentials to be provided.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.ClientAuthError`
          :exc:`~zhmcclient.ServerAuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        if self._userid is None:
            raise ClientAuthError("Userid is not provided.")

        # Determine working HMC for this session
        self._actual_host = self._determine_actual_host()
        self._base_url = self._create_base_url(self._actual_host, self._port)

        if self._password is None:
            if self._get_password:
                self._password = \
                    self._get_password(self._actual_host, self._userid)
            else:
                raise ClientAuthError("Password is not provided.")

        # Create an HMC session
        logon_uri = '/api/sessions'
        logon_body = {
            'userid': self._userid,
            'password': self._password
        }
        self._headers.pop('X-API-Session', None)  # Just in case
        self._session = self._new_session(self.retry_timeout_config)
        logon_res = self.post(logon_uri, body=logon_body, logon_required=False)
        self._session_id = logon_res['api-session']
        self._session_credential = logon_res['session-credential']
        self._headers['X-API-Session'] = self._session_id
        self._object_topic = logon_res['notification-topic']
        self._job_topic = logon_res['job-notification-topic']

    @staticmethod
    def _create_base_url(host, port):
        """
        Encapsulates how the base URL of the HMC is constructed.
        """
        return f"{_HMC_SCHEME}://{host}:{port}"

    def _determine_actual_host(self):
        """
        Determine the actual HMC host to be used.

        If a single HMC host is specified, that host is used without further
        verification as to whether it is available.

        If more than one HMC host is specified, the first available host is
        used. Availability of the HMC is determined using the
        'Query API Version' operation, for which no logon is required.
        If no available HMC can be found, raises the ConnectionError of the
        last HMC that was tried.
        """

        if len(self._hosts) == 1:
            host = self._hosts[0]
            HMC_LOGGER.debug("Using the only HMC specified without verifying "
                             "its availability: %s", host)
            return host

        last_exc = None
        for host in self._hosts:
            HMC_LOGGER.debug("Trying HMC for availability: %s", host)
            self._base_url = self._create_base_url(host, self._port)
            try:
                self.get('/api/version', logon_required=False,
                         renew_session=False)
            except ConnectionError as exc:
                last_exc = exc
                continue

            HMC_LOGGER.debug("Using available HMC: %s", host)
            return host

        HMC_LOGGER.debug("Did not find an available HMC in: %s",
                         self._hosts)
        raise last_exc

    @staticmethod
    def _new_session(retry_timeout_config):
        """
        Return a new `requests.Session` object.
        """
        retry = urllib3.Retry(
            total=retry_timeout_config.connect_retries,
            connect=retry_timeout_config.connect_retries,
            read=retry_timeout_config.read_retries,
            allowed_methods=retry_timeout_config.allowed_methods,
            redirect=retry_timeout_config.max_redirects)
        session = requests.Session()
        session.mount('https://',
                      requests.adapters.HTTPAdapter(max_retries=retry))
        session.mount('http://',
                      requests.adapters.HTTPAdapter(max_retries=retry))
        return session

    def _do_logoff(self):
        """
        Log off, unconditionally.

        This deletes the session on the HMC. If that deletion operation fails
        due to an invalid session ID, that failure is ignored.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.ConnectionError`
        """
        session_uri = '/api/sessions/this-session'
        try:
            self.delete(session_uri, logon_required=False, renew_session=False)
        except (ServerAuthError, ConnectionError):
            # HMC shutdown or broken network causes ConnectionError.
            # Invalid credentials cause ServerAuthError.
            pass

        self._actual_host = None
        self._base_url = None
        self._session_id = None
        self._session = None
        self._headers.pop('X-API-Session', None)
        self._object_topic = None
        self._job_topic = None

    @staticmethod
    def _log_http_request(
            method, url, resource, headers=None, content=None,
            content_len=None):
        """
        Log the HTTP request of an HMC REST API call, at the debug level.

        Parameters:

          method (:term:`string`): HTTP method name in upper case, e.g. 'GET'

          url (:term:`string`): HTTP URL (base URL and operation URI)

          headers (iterable): HTTP headers used for the request

          content (:term:`string`): HTTP body (aka content) used for the
            request (byte string or unicode string)

          content_len (int): Length of content in Bytes, or `None` for
            determining the length from the content string
        """

        content_msg = None
        if content is not None:
            if isinstance(content, bytes):
                content = content.decode('utf-8', errors='ignore')
            assert isinstance(content, str)
            if content_len is None:
                content_len = len(content)  # may change after JSON conversion
            try:
                content_dict = json2dict(content)
            except ValueError:
                # If the content is not JSON, we assume it does not contain
                # structured data such as a password or session IDs.
                pass
            else:
                for prop in BLANKED_OUT_PROPERTIES:
                    try:
                        content_dict[prop] = BLANKED_OUT_STRING
                    except KeyError:
                        pass
                content = dict2json(content_dict)
            trunc = 30000
            if content_len > trunc:
                content_label = f"content(first {trunc} B of {content_len} B)"
                content_msg = content[0:trunc] + '...(truncated)'
            else:
                content_label = f'content({content_len} B)'
                content_msg = content
        else:
            content_label = 'content'
            content_msg = content

        if resource:
            names = []
            res_class = resource.manager.class_name
            while resource:
                # Using resource.name gets into an infinite recursion when
                # the resource name is not present, due to pulling the
                # properties in that case. We take the careful approach.
                name_prop = resource.manager.name_prop
                name = resource.properties.get(name_prop, '<unknown>')
                names.insert(0, name)
                resource = resource.manager.parent
            res_str = f" ({res_class} {'.'.join(names)})"
        else:
            res_str = ""

        HMC_LOGGER.debug("Request: %s %s%s, headers: %r, %s: %r",
                         method, url, res_str, _headers_for_logging(headers),
                         content_label, content_msg)

    @staticmethod
    def _log_http_response(
            method, url, resource, status, headers=None, content=None):
        """
        Log the HTTP response of an HMC REST API call, at the debug level.

        Parameters:

          method (:term:`string`): HTTP method name in upper case, e.g. 'GET'

          url (:term:`string`): HTTP URL (base URL and operation URI)

          status (integer): HTTP status code

          headers (iterable): HTTP headers returned in the response

          content (:term:`string`): HTTP body (aka content) returned in the
            response (byte string or unicode string)
        """

        if content is not None:
            if isinstance(content, bytes):
                content = content.decode('utf-8')
            assert isinstance(content, str)
            content_len = len(content)  # may change after JSON conversion
            try:
                content_dict = json2dict(content)
            except ValueError:
                # If the content is not JSON (e.g. response from metrics
                # context retrieval), we assume it does not contain structured
                # data such as a password or session IDs.
                pass
            else:
                if 'request-headers' in content_dict:
                    headers_dict = content_dict['request-headers']
                    if 'x-api-session' in headers_dict:
                        headers_dict['x-api-session'] = BLANKED_OUT_STRING
                if 'api-session' in content_dict:
                    content_dict['api-session'] = BLANKED_OUT_STRING
                if 'session-credential' in content_dict:
                    content_dict['session-credential'] = BLANKED_OUT_STRING
                content = dict2json(content_dict)
            if status >= 400:
                content_label = 'content'
                content_msg = content
            else:
                trunc = 30000
                if content_len > trunc:
                    content_label = \
                        f"content(first {trunc} B of {content_len} B)"
                    content_msg = content[0:trunc] + '...(truncated)'
                else:
                    content_label = f'content({len(content)} B)'
                    content_msg = content
        else:
            content_label = 'content'
            content_msg = content

        if resource:
            names = []
            res_class = resource.manager.class_name
            while resource:
                # Using resource.name gets into an infinite recursion when
                # the resource name is not present, due to pulling the
                # properties in that case. We take the careful approach.
                name_prop = resource.manager.name_prop
                name = resource.properties.get(name_prop, '<unknown>')
                names.insert(0, name)
                resource = resource.manager.parent
            res_str = f" ({res_class} {'.'.join(names)})"
        else:
            res_str = ""

        HMC_LOGGER.debug("Respons: %s %s%s, status: %s, "
                         "headers: %r, %s: %r",
                         method, url, res_str, status,
                         _headers_for_logging(headers),
                         content_label, content_msg)

    @logged_api_call
    def get(self, uri, resource=None, logon_required=True, renew_session=True):
        """
        Perform the HTTP GET method against the resource identified by a URI.

        A set of standard HTTP headers is automatically part of the request.

        If the HMC session token is expired, this method re-logs on and retries
        the operation.

        Parameters:

          uri (:term:`string`):
            Relative URI path of the resource, e.g. "/api/session".
            This URI is relative to the base URL of the session (see
            the :attr:`~zhmcclient.Session.base_url` property).
            Must not be `None`.

          logon_required (bool):
            Boolean indicating whether the operation requires that the session
            is logged on to the HMC. For example, the API version retrieval
            operation does not require that.

          renew_session (bool):
            Boolean indicating whether the session should be renewed in case
            it is expired.

        Returns:

          :term:`json object` with the operation result.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.ClientAuthError`
          :exc:`~zhmcclient.ServerAuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        if logon_required:
            self.logon()
        elif self._base_url is None:
            self._actual_host = self._determine_actual_host()
            self._base_url = \
                self._create_base_url(self._actual_host, self._port)
        url = self._base_url + uri
        self._log_http_request('GET', url, resource=resource,
                               headers=self.headers)
        stats = self.time_stats_keeper.get_stats('get ' + uri)
        stats.begin()
        req = self._session or requests
        req_timeout = (self.retry_timeout_config.connect_timeout,
                       self.retry_timeout_config.read_timeout)
        try:
            result = req.get(url, headers=self.headers, verify=self.verify_cert,
                             timeout=req_timeout)
        # Note: The requests method may raise OSError/IOError in case of
        # HMC certificate validation issues (e.g. incorrect cert path)
        except (requests.exceptions.RequestException, OSError) as exc:
            _handle_request_exc(exc, self.retry_timeout_config)
        finally:
            stats.end()
        self._log_http_response('GET', url, resource=resource,
                                status=result.status_code,
                                headers=result.headers,
                                content=result.content)

        if result.status_code == 200:
            return _result_object(result)
        if result.status_code == 403:
            result_object = _result_object(result)
            reason = result_object.get('reason', None)
            message = result_object.get('message', None)
            HMC_LOGGER.debug("Received HTTP status 403.%d on GET %s: %s",
                             reason, uri, message)

            if reason in (4, 5):
                # 403.4: No session ID was provided
                # 403.5: Session ID was invalid
                if renew_session:
                    self._do_logon()
                    return self.get(
                        uri, resource=resource, logon_required=False,
                        renew_session=False)

            if reason == 1:
                # Login user's authentication is fine; this is an authorization
                # issue, so we don't raise ServerAuthError.
                raise HTTPError(result_object)

            msg = result_object.get('message', None)
            raise ServerAuthError(
                "HTTP authentication failed with "
                f"{result.status_code},{reason}: {msg}",
                HTTPError(result_object))

        result_object = _result_object(result)
        raise HTTPError(result_object)

    @logged_api_call
    def post(self, uri, resource=None, body=None, logon_required=True,
             wait_for_completion=False, operation_timeout=None,
             renew_session=True):
        """
        Perform the HTTP POST method against the resource identified by a URI,
        using a provided request body.

        A set of standard HTTP headers is automatically part of the request.

        HMC operations using HTTP POST are either synchronous or asynchronous.
        Asynchronous operations return the URI of an asynchronously executing
        job that can be queried for status and result.

        Examples for synchronous operations:

        * With no result: "Logon", "Update CPC Properties"
        * With a result: "Create Partition"

        Examples for asynchronous operations:

        * With no result: "Start Partition"

        The `wait_for_completion` parameter of this method can be used to deal
        with asynchronous HMC operations in a synchronous way.

        If executing the operation reveals that the HMC session token is
        expired, this method re-logs on and retries the operation.

        The timeout and retry

        Parameters:

          uri (:term:`string`):
            Relative URI path of the resource, e.g. "/api/session".
            This URI is relative to the base URL of the session (see the
            :attr:`~zhmcclient.Session.base_url` property).
            Must not be `None`.

          body (:term:`json object` or :term:`string` or file-like object):
            The HTTP request body (payload).
            If a JSON object (=dict) is provided, it will be serialized into
            a UTF-8 encoded binary string.
            If a Unicode string is provided, it will be encoded into a UTF-8
            encoded binary string.
            If a binary string is provided, it will be used unchanged.
            If a file-like object is provided, it must return binary strings,
            i.e. the file must have been opened in binary mode.
            `None` means the same as an empty dictionary, namely that no HTTP
            body is included in the request.

          logon_required (bool):
            Boolean indicating whether the operation requires that the session
            is logged on to the HMC. For example, the "Logon" operation does
            not require that.

          wait_for_completion (bool):
            Boolean controlling whether this method should wait for completion
            of the requested asynchronous HMC operation.

            A value of `True` will cause an additional entry in the time
            statistics to be created that represents the entire asynchronous
            operation including the waiting for its completion.
            That time statistics entry will have a URI that is the targeted
            URI, appended with "+completion".

            For synchronous HMC operations, this parameter has no effect on
            the operation execution or on the return value of this method, but
            it should still be set (or defaulted) to `False` in order to avoid
            the additional entry in the time statistics.

          operation_timeout (:term:`number`):
            Timeout in seconds, when waiting for completion of an asynchronous
            operation. The special value 0 means that no timeout is set. `None`
            means that the default async operation timeout of the session is
            used.

            For `wait_for_completion=True`, a
            :exc:`~zhmcclient.OperationTimeout` is raised when the timeout
            expires.

            For `wait_for_completion=False`, this parameter has no effect.

          renew_session (bool):
            Boolean indicating whether the session should be renewed in case
            it is expired.

        Returns:

          : A :term:`json object` or `None` or a :class:`~zhmcclient.Job`
          object, as follows:

          * For synchronous HMC operations, and for asynchronous HMC
            operations with `wait_for_completion=True`:

            If this method returns, the HMC operation has completed
            successfully (otherwise, an exception is raised).
            For asynchronous HMC operations, the associated job has been
            deleted.

            The return value is the result of the HMC operation as a
            :term:`json object`, or `None` if the operation has no result.
            See the section in the :term:`HMC API` book about the specific
            HMC operation for a description of the members of the returned
            JSON object.

          * For asynchronous HMC operations with `wait_for_completion=False`:

            If this method returns, the asynchronous execution of the HMC
            operation has been started successfully as a job on the HMC (if
            the operation could not be started, an exception is raised).

            The return value is a :class:`~zhmcclient.Job` object
            representing the job on the HMC.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.ClientAuthError`
          :exc:`~zhmcclient.ServerAuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for completion of the asynchronous operation.
          :exc:`TypeError`: Body has invalid type.
        """
        if logon_required:
            self.logon()
        elif self._base_url is None:
            self._actual_host = self._determine_actual_host()
            self._base_url = \
                self._create_base_url(self._actual_host, self._port)
        url = self._base_url + uri
        headers = self.headers.copy()  # Standard headers

        log_len = None
        if body is None:
            data = None
            log_data = None
        elif isinstance(body, dict):
            data = json.dumps(body)
            # Produces unicode string on py3, and unicode or byte string on py2.
            # Content-type is already set to 'application/json' in standard
            # headers.
            if isinstance(data, str):
                log_data = data
                data = data.encode('utf-8')
            else:
                log_data = data
        elif isinstance(body, str):
            data = body.encode('utf-8')
            log_data = body
            headers['Content-type'] = 'application/octet-stream'
        elif isinstance(body, bytes):
            data = body
            log_data = body
            headers['Content-type'] = 'application/octet-stream'
        elif isinstance(body, Iterable):
            # File-like objects, e.g. io.BufferedReader or io.TextIOWrapper
            # returned from open() or io.open().
            data = body
            try:
                mode = body.mode
            except AttributeError:
                mode = 'unknown'
            log_data = f"<file-like object with mode {mode}>"
            log_len = -1
            headers['Content-type'] = 'application/octet-stream'
        else:
            raise TypeError(f"Body has invalid type: {type(body)}")

        self._log_http_request('POST', url, resource=resource, headers=headers,
                               content=log_data, content_len=log_len)
        req = self._session or requests
        req_timeout = (self.retry_timeout_config.connect_timeout,
                       self.retry_timeout_config.read_timeout)
        if wait_for_completion:
            stats_total = self.time_stats_keeper.get_stats(
                'post ' + uri + '+completion')
            stats_total.begin()
        try:
            stats = self.time_stats_keeper.get_stats('post ' + uri)
            stats.begin()
            try:
                if data is None:
                    result = req.post(url, headers=headers,
                                      verify=self.verify_cert,
                                      timeout=req_timeout)
                else:
                    result = req.post(url, data=data, headers=headers,
                                      verify=self.verify_cert,
                                      timeout=req_timeout)
            # Note: The requests method may raise OSError/IOError in case of
            # HMC certificate validation issues (e.g. incorrect cert path)
            except (requests.exceptions.RequestException, OSError) \
                    as exc:
                _handle_request_exc(exc, self.retry_timeout_config)
            finally:
                stats.end()
            self._log_http_response('POST', url, resource=resource,
                                    status=result.status_code,
                                    headers=result.headers,
                                    content=result.content)

            if result.status_code in (200, 201):
                return _result_object(result)

            if result.status_code == 204:
                # No content
                return None

            if result.status_code == 202:
                if result.content == b'':
                    # Some operations (e.g. "Restart Console",
                    # "Shutdown Console" or "Cancel Job") return 202
                    # with no response content.
                    return None

                # This is the most common case to return 202: An
                # asynchronous job has been started.
                result_object = _result_object(result)
                job_uri = result_object['job-uri']
                job = Job(self, job_uri, 'POST', uri)
                if wait_for_completion:
                    return job.wait_for_completion(operation_timeout)
                return job

            if result.status_code == 403:
                result_object = _result_object(result)
                reason = result_object.get('reason', None)
                message = result_object.get('message', None)
                HMC_LOGGER.debug("Received HTTP status 403.%d on GET %s: %s",
                                 reason, uri, message)

                if reason in (4, 5):
                    # 403.4: No session ID was provided
                    # 403.5: Session ID was invalid
                    if renew_session:
                        self._do_logon()
                        return self.post(
                            uri, resource=resource, body=body,
                            logon_required=False, renew_session=False,
                            wait_for_completion=wait_for_completion,
                            operation_timeout=operation_timeout)

                if reason == 1:
                    # Login user's authentication is fine; this is an
                    # authorization issue, so we don't raise ServerAuthError.
                    raise HTTPError(result_object)

                msg = result_object.get('message', None)
                raise ServerAuthError(
                    "HTTP authentication failed with "
                    f"{result.status_code},{reason}: {msg}",
                    HTTPError(result_object))

            result_object = _result_object(result)
            raise HTTPError(result_object)

        finally:
            if wait_for_completion:
                stats_total.end()

    @logged_api_call
    def delete(
            self, uri, resource=None, logon_required=True, renew_session=True):
        """
        Perform the HTTP DELETE method against the resource identified by a
        URI.

        A set of standard HTTP headers is automatically part of the request.

        If the HMC session token is expired, this method re-logs on and retries
        the operation.

        Parameters:

          uri (:term:`string`):
            Relative URI path of the resource, e.g.
            "/api/session/{session-id}".
            This URI is relative to the base URL of the session (see
            the :attr:`~zhmcclient.Session.base_url` property).
            Must not be `None`.

          logon_required (bool):
            Boolean indicating whether the operation requires that the session
            is logged on to the HMC. For example, for the logoff operation, it
            does not make sense to first log on.

          renew_session (bool):
            Boolean indicating whether the session should be renewed in case
            it is expired. For example, for the logoff operation, it does not
            make sense to do that.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.ClientAuthError`
          :exc:`~zhmcclient.ServerAuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        if logon_required:
            self.logon()
        elif self._base_url is None:
            self._actual_host = self._determine_actual_host()
            self._base_url = \
                self._create_base_url(self._actual_host, self._port)
        url = self._base_url + uri
        self._log_http_request('DELETE', url, resource=resource,
                               headers=self.headers)
        stats = self.time_stats_keeper.get_stats('delete ' + uri)
        stats.begin()
        req = self._session or requests
        req_timeout = (self.retry_timeout_config.connect_timeout,
                       self.retry_timeout_config.read_timeout)
        try:
            result = req.delete(url, headers=self.headers,
                                verify=self.verify_cert, timeout=req_timeout)
        # Note: The requests method may raise OSError/IOError in case of
        # HMC certificate validation issues (e.g. incorrect cert path)
        except (requests.exceptions.RequestException, OSError) as exc:
            _handle_request_exc(exc, self.retry_timeout_config)
        finally:
            stats.end()
        self._log_http_response('DELETE', url, resource=resource,
                                status=result.status_code,
                                headers=result.headers,
                                content=result.content)

        if result.status_code in (200, 204):
            return

        if result.status_code == 403:
            result_object = _result_object(result)
            reason = result_object.get('reason', None)
            message = result_object.get('message', None)
            HMC_LOGGER.debug("Received HTTP status 403.%d on GET %s: %s",
                             reason, uri, message)

            if reason in (4, 5):
                # 403.4: No session ID was provided
                # 403.5: Session ID was invalid
                if renew_session:
                    self._do_logon()
                    self.delete(uri, resource=resource, logon_required=False,
                                renew_session=False)
                    return

            if reason == 1:
                # Login user's authentication is fine; this is an authorization
                # issue, so we don't raise ServerAuthError.
                raise HTTPError(result_object)

            msg = result_object.get('message', None)
            raise ServerAuthError(
                "HTTP authentication failed with "
                f"{result.status_code},{reason}: {msg}",
                HTTPError(result_object))

        result_object = _result_object(result)
        raise HTTPError(result_object)

    @logged_api_call
    def get_notification_topics(self):
        """
        The 'Get Notification Topics' operation returns a structure that
        describes the JMS notification topics associated with this session.

        Returns:

          : A list with one item for each notification topic. Each item is a
          dictionary with the following keys:

          * ``"topic-type"`` (string): Topic type, e.g. "job-notification".
          * ``"topic-name"`` (string): Topic name; can be used for
            subscriptions.
          * ``"object-uri"`` (string): When topic-type is
            "os-message-notification", this item is the canonical URI path
            of the Partition for which this topic exists.
            This field does not exist for the other topic types.
          * ``"include-refresh-messages"`` (bool): When the topic-type is
            "os-message-notification", this item indicates whether refresh
            operating system messages will be sent on this topic.
        """
        topics_uri = '/api/sessions/operations/get-notification-topics'
        response = self.get(topics_uri)
        return response['topics']

    def auto_update_subscribed(self):
        """
        Return whether this session is currently subscribed for
        :ref:`auto-updating`.

        Return:
          bool: Indicates whether session is subscribed.
        """
        return self._auto_updater.is_open()

    @logged_api_call
    def subscribe_auto_update(self):
        """
        Subscribe this session for :ref:`auto-updating`, if not currently
        subscribed.

        When not yet subscribed, the session is also logged on.

        When subscribed, object notifications will be sent by the HMC as
        resource objects on the HMC change their properties or come or go.
        These object notifications will be received by the client and will then
        update the properties of any Python resource objects that are enabled
        for auto-updating.

        This method is automatically called by
        :meth:`~zhmcclient.BaseResource.enable_auto_update` and thus does not
        need to be called by the user.
        """
        if not self._auto_updater.is_open():
            self._auto_updater.open()

    @logged_api_call
    def unsubscribe_auto_update(self):
        """
        Unsubscribe this session from :ref:`auto-updating`, if
        currently subscribed.

        When unsubscribed, object notifications are no longer sent by the HMC.

        This method is automatically called by
        :meth:`~zhmcclient.BaseResource.disable_auto_update` and thus does not
        need to be called by the user.
        """
        if self._auto_updater.is_open():
            self._auto_updater.close()


class Job:
    """
    A job on the HMC that performs an asynchronous HMC operation.

    This class supports checking the job for completion, and waiting for job
    completion.
    """

    def __init__(self, session, uri, op_method, op_uri):
        """
        Parameters:

          session (:class:`~zhmcclient.Session`):
            Session with the HMC.
            Must not be `None`.

          uri (:term:`string`):
            Canonical URI of the job on the HMC.
            Must not be `None`.

            Example: ``"/api/jobs/{job-id}"``

          op_method (:term:`string`):
            Name of the HTTP method of the operation that is executing
            asynchronously on the HMC.
            Must not be `None`.

            Example: ``"POST"``

          op_uri (:term:`string`):
            Canonical URI of the operation that is executing asynchronously on
            the HMC.
            Must not be `None`.

            Example: ``"/api/partitions/{partition-id}/stop"``
        """
        self._session = session
        self._uri = uri
        self._op_method = op_method
        self._op_uri = op_uri

    @property
    def session(self):
        """
        :class:`~zhmcclient.Session`: Session with the HMC.
        """
        return self._session

    @property
    def uri(self):
        """
        :term:`string`: Canonical URI of the job on the HMC.

        Example: ``"/api/jobs/{job-id}"``
        """
        return self._uri

    @property
    def op_method(self):
        """
        :term:`string`: Name of the HTTP method of the operation that is
        executing asynchronously on the HMC.

        Example: ``"POST"``
        """
        return self._op_method

    @property
    def op_uri(self):
        """
        :term:`string`: Canonical URI of the operation that is executing
        asynchronously on the HMC.

        Example: ``"/api/partitions/{partition-id}/stop"``
        """
        return self._op_uri

    @logged_api_call
    def query_status(self):
        """
        Get the current status of this job, and if completed also the
        operation results.

        This method performs the "Query Job Status" operation on the job.

        This is a low level operation, consider using
        :meth:`~zhmcclient.Job.check_for_completion` or
        :meth:`~zhmcclient.Job.wait_for_completion` instead.

        If the job no longer exists, :exc:`~zhmcclient.HTTPError` is raised
        with status code 404 and reason code 1.

        Returns:

          tuple(job_status, op_status, op_reason, op_result): With the following
          items:

          * job_status(string): Job status, one of:

            - "running" - indicates that the job was found and it has not ended
              at the time of the query.
            - "cancel-pending" - indicates that the job was found and it has
              not ended but cancellation has been requested.
            - "canceled" - indicates that the job's normal course of execution
              was interrupted by a cancel request, and the job has now ended.
            - "complete" - indicates that the job was found and has completed
              the operation, and the job has now ended.

          * op_status(int): HTTP status code of the operation performed by
            the job. Will be `None` if the job has not ended.

          * op_reason(int): HTTP reason code of the operation performed by
            the job. Will be `None` if the job has not ended.

          * op_result(dict): Result of the operation performed by
            the job, as described in the zhmcclient method that performed
            the operation. Will be `None` if the job has not ended.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.ClientAuthError`
          :exc:`~zhmcclient.ServerAuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        try:
            result = self.session.get(self.uri)
        except Error as exc:
            HMC_LOGGER.debug("Request: GET %s failed with %s: %s",
                             self.uri, exc.__class__.__name__, exc)
            raise

        job_status = result['status']
        op_status = result.get('job-status-code', None)
        op_reason = result.get('job-reason-code', None)
        op_result = result.get('job-results', None)

        return job_status, op_status, op_reason, op_result

    @logged_api_call
    def delete(self):
        """
        Delete this ended job on the HMC.

        This method performs the "Delete Completed Job Status" operation on the
        job.

        This is a low level operation, consider using
        :meth:`~zhmcclient.Job.check_for_completion` or
        :meth:`~zhmcclient.Job.wait_for_completion` instead, which delete the
        ended job.

        If the job has not ended (i.e. its `status` property is not "canceled"
        or"complete"), :exc:`~zhmcclient.HTTPError` is raised with status code
        409 and reason code 40.

        If the job no longer exists, :exc:`~zhmcclient.HTTPError` is raised
        with status code 404 and reason code 1.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.ClientAuthError`
          :exc:`~zhmcclient.ServerAuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        try:
            self.session.delete(self.uri)
        except Error as exc:
            HMC_LOGGER.debug("Request: DELETE %s failed with %s: %s",
                             self.uri, exc.__class__.__name__, exc)
            raise

    @logged_api_call
    def check_for_completion(self):
        """
        Check once for completion of the job and return completion status and
        result if it has completed.

        If the job completed in error, an :exc:`~zhmcclient.HTTPError`
        exception is raised.

        Returns:

          : A tuple (status, result) with:

          * status (:term:`string`): Completion status of the job, as
            returned in the ``status`` field of the response body of the
            "Query Job Status" HMC operation, as follows:

            * ``"complete"``: Job completed (successfully).
            * any other value: Job is not yet complete.

          * result (:term:`json object` or `None`): `None` for incomplete
            jobs. For completed jobs, the result of the original asynchronous
            operation that was performed by the job, from the ``job-results``
            field of the response body of the "Query Job Status" HMC
            operation. That result is a :term:`json object` as described
            for the asynchronous operation, or `None` if the operation has no
            result.

        Raises:

          :exc:`~zhmcclient.HTTPError`: The job completed in error, or the job
            status cannot be retrieved, or the job cannot be deleted.
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.ClientAuthError`
          :exc:`~zhmcclient.ServerAuthError`
          :exc:`~zhmcclient.ConnectionError`
        """

        try:
            job_result_obj = self.session.get(self.uri)
        except Error as exc:
            HMC_LOGGER.debug("Request: GET %s failed with %s: %s",
                             self.uri, exc.__class__.__name__, exc)
            raise

        job_status = job_result_obj['status']
        if job_status == 'complete':
            self.session.delete(self.uri)
            op_status_code = job_result_obj['job-status-code']
            if op_status_code in (200, 201):
                op_result_obj = job_result_obj.get('job-results', None)
            elif op_status_code == 204:
                # No content
                op_result_obj = None
            else:
                error_result_obj = job_result_obj.get('job-results', None)
                if not error_result_obj:
                    message = None
                elif 'message' in error_result_obj:
                    message = error_result_obj['message']
                elif 'error' in error_result_obj:
                    message = error_result_obj['error']
                else:
                    message = None
                error_obj = {
                    'http-status': op_status_code,
                    'reason': job_result_obj['job-reason-code'],
                    'message': message,
                    'request-method': self.op_method,
                    'request-uri': self.op_uri,
                }
                raise HTTPError(error_obj)
        else:
            op_result_obj = None
        return job_status, op_result_obj

    @logged_api_call
    def wait_for_completion(self, operation_timeout=None):
        """
        Wait for completion of the job, then delete the job on the HMC and
        return the result of the original asynchronous HMC operation, if it
        completed successfully.

        If the job completed in error, an :exc:`~zhmcclient.HTTPError`
        exception is raised.

        Parameters:

          operation_timeout (:term:`number`):
            Timeout in seconds, when waiting for completion of the job. The
            special value 0 means that no timeout is set. `None` means that the
            default async operation timeout of the session is used.

            If the timeout expires, a :exc:`~zhmcclient.OperationTimeout`
            is raised.

            This method gives completion of the job priority over strictly
            achieving the timeout. This may cause a slightly longer duration of
            the method than prescribed by the timeout.

        Returns:

          :term:`json object` or `None`:
            The result of the original asynchronous operation that was
            performed by the job, from the ``job-results`` field of the
            response body of the "Query Job Status" HMC operation. That result
            is a :term:`json object` as described for the asynchronous
            operation, or `None` if the operation has no result.

        Raises:

          :exc:`~zhmcclient.HTTPError`: The job completed in error, or the job
            status cannot be retrieved, or the job cannot be deleted.
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.ClientAuthError`
          :exc:`~zhmcclient.ServerAuthError`
          :exc:`~zhmcclient.OperationTimeout`: The timeout expired while
            waiting for job completion.
        """

        if operation_timeout is None:
            operation_timeout = \
                self.session.retry_timeout_config.operation_timeout
        if operation_timeout > 0:
            start_time = time.time()

        while True:
            try:
                job_status, op_result_obj = self.check_for_completion()
            except ConnectionError:
                HMC_LOGGER.debug("Retrying after ConnectionError while waiting"
                                 " for completion of job %s. This could be "
                                 "because HMC is restarting.", self.uri)
                job_status = None

            # We give completion of status priority over strictly achieving
            # the timeout, so we check status first. This may cause a longer
            # duration of the method than prescribed by the timeout.
            if job_status == 'complete':
                return op_result_obj

            if operation_timeout > 0:
                current_time = time.time()
                if current_time > start_time + operation_timeout:
                    raise OperationTimeout(
                        f"Waiting for completion of job {self.uri} timed out "
                        f"(operation timeout: {operation_timeout} s)",
                        operation_timeout)

            time.sleep(10)  # Avoid hot spin loop

    @logged_api_call
    def cancel(self):
        """
        Attempt to cancel this job.

        This method performs the "Cancel Job" operation on the job.

        The specific nature of the job and its current state of execution can
        affect the success of the cancellation.

        Not all jobs support cancellation; this is described in each zhmcclient
        method that can return a job.

        If the job exists, supports cancellation and has not yet completed (i.e.
        its `status` property is "running"), the cancellation is made pending
        for the job and its `status` property is changed to "cancel-pending".

        If the operation performed by the job does not support cancellation,
        :exc:`~zhmcclient.HTTPError` is raised with status code 404 and reason
        code 4.

        If the job supports cancellation and exists, but already has a
        cancellation request pending (i.e. its `status` property is
        "cancel-pending"), :exc:`~zhmcclient.HTTPError` is raised with status
        409 and reason code 42.

        If the job supports cancellation and exists, but already has ended
        (i.e. its `status` property is "complete" or "canceled"),
        :exc:`~zhmcclient.HTTPError` is raised with status code 409 and reason
        code 41.

        If the job supports cancellation but no longer exists,
        :exc:`~zhmcclient.HTTPError` is raised with status code 404 and reason
        code 1.

        Raises:

          :exc:`~zhmcclient.HTTPError`: The job cancellation attempt failed.
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.ClientAuthError`
          :exc:`~zhmcclient.ServerAuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        uri = f'{self.uri}/operations/cancel'
        try:
            self.session.post(uri)
        except Error as exc:
            HMC_LOGGER.debug("Request: POST %s failed with %s: %s",
                             uri, exc.__class__.__name__, exc)
            raise


def _text_repr(text, max_len=1000):
    """
    Return the input text as a Python string representation (i.e. using repr())
    that is limited to a maximum length.
    """
    if text is None:
        text_repr = 'None'
    elif len(text) > max_len:
        text_repr = repr(text[0:max_len]) + '...'
    else:
        text_repr = repr(text)
    return text_repr


def _result_object(result):
    """
    Return the JSON payload in the HTTP response as a Python dict.

    Parameters:

        result (requests.Response): HTTP response object.

    Raises:

        :exc:`~zhmcclient.ParseError`: Error parsing the returned JSON.
    """
    content_type = result.headers.get('content-type', None)

    if content_type is None or content_type.startswith('application/json'):
        # This function is only called when there is content expected.
        # Therefore, a response without content will result in a ParseError.
        try:
            return result.json(object_pairs_hook=OrderedDict)
        except ValueError as exc:
            new_exc = ParseError(
                f"JSON parse error in HTTP response: {exc.args[0]}. "
                f"HTTP request: {result.request.method} {result.request.url}. "
                f"Response status {result.status_code}. "
                f"Response content-type: {content_type!r}. "
                f"Content (max.1000, decoded using {result.encoding}): "
                f"{_text_repr(result.text, 1000)}")
            new_exc.__cause__ = None
            raise new_exc  # zhmcclient.ParseError

    if content_type.startswith('text/html'):
        # We are in some error situation. The HMC returns HTML content
        # for some 5xx status codes. We try to deal with it somehow,
        # but we are not going as far as real HTML parsing.
        m = re.search(r'charset=([^;,]+)', content_type)
        if m:
            encoding = m.group(1)  # e.g. RFC "ISO-8859-1"
        else:
            encoding = 'utf-8'
        try:
            html_uni = result.content.decode(encoding)
        except LookupError:
            html_uni = result.content.decode()

        # We convert to one line to be regexp-friendly.
        html_oneline = html_uni.replace('\r\n', '\\n').replace('\r', '\\n').\
            replace('\n', '\\n')

        # Check for some well-known errors:
        if re.search(r'javax\.servlet\.ServletException: '
                     r'Web Services are not enabled\.', html_oneline):
            html_title = "Console Configuration Error"
            html_details = "Web Services API is not enabled on the HMC."
            html_reason = HTML_REASON_WEB_SERVICES_DISABLED
        else:
            m = re.search(
                r'<title>([^<]*)</title>.*'
                r'<h2>Details:</h2>(.*)(<hr size="1" noshade>)?</body>',
                html_oneline)
            if m:
                html_title = m.group(1)
                # Spend a reasonable effort to make the HTML readable:
                html_details = m.group(2).replace('<p>', '\\n').\
                    replace('<br>', '\\n').replace('\\n\\n', '\\n').strip()
            else:
                html_title = "Console Internal Error"
                html_details = f"Response body: {html_uni!r}"
            html_reason = HTML_REASON_OTHER
        message = f"{html_title}: {html_details}"

        # We create a minimal JSON error object (to the extent we use it
        # when processing it):
        result_obj = {
            'http-status': result.status_code,
            'reason': html_reason,
            'message': message,
            'request-uri': result.request.url,
            'request-method': result.request.method,
        }
        return result_obj

    if content_type.startswith('application/vnd.ibm-z-zmanager-metrics'):
        content_bytes = result.content
        assert isinstance(content_bytes, bytes)
        return content_bytes.decode('utf-8')  # as a unicode object

    raise ParseError(
        f"Unknown content type in HTTP response: {content_type}. "
        f"HTTP request: {result.request.method} {result.request.url}. "
        f"Response status {result.status_code}. "
        f"Content (max.1000, decoded using {result.encoding}): "
        f"{_text_repr(result.text, 1000)}")


def json2dict(json_str):
    """
    Convert a JSON string into a dict.

    Parameters:

      json_str (string): Unicode or binary string in JSON format.

    Returns:

      dict: JSON string converted to a dict.

    Raises:

      ValueError: Cannot parse JSON string
    """
    json_dict = json.loads(json_str)  # May raise ValueError
    return json_dict


def dict2json(json_dict):
    """
    Convert a dict into a JSON string.

    Parameters:

      json_dict (dict): The dict.

    Returns:

      unicode string (py3) or byte string (py2): Dict converted to a JSON
      string.
    """
    json_str = json.dumps(json_dict)
    return json_str



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Base definitions for resource classes.

Resource objects represent the real manageable resources in the systems managed
by the HMC.
"""

import time
import threading
from collections import OrderedDict
# import contextlib
from immutable_views import DictView

from ._logging import logged_api_call
from ._utils import repr_dict, repr_timestamp, repr_obj_id
from ._exceptions import CeasedExistence, HTTPError

__all__ = ['BaseResource']


class BaseResource:
    """
    Abstract base class for resource classes (e.g. :class:`~zhmcclient.Cpc`)
    representing manageable resources.

    It defines the interface for the derived resource classes, and implements
    methods that have a common implementation for the derived resource classes.

    Objects of derived resource classes are representations of the actual
    manageable resources in the HMC or in systems managed by the HMC.

    Objects of derived resource classes should not be created by users of this
    package by simply instantiating the derived resource classes. Instead, such
    objects are created by this package and are returned to the user as a
    result of methods such as :meth:`~zhmcclient.BaseManager.find` or
    :meth:`~zhmcclient.BaseManager.list`. For this reason, the `__init__()`
    method of this class and of its derived resource classes are considered
    internal interfaces and their parameters are not documented and may change
    incompatibly.
    """

    def __init__(self, manager, uri, name, properties):
        # This method intentionally has no docstring, because it is internal.
        #
        # Parameters:
        #   manager (subclass of :class:`~zhmcclient.BaseManager`):
        #     Manager object for this resource object (and for all resource
        #     objects of the same type in the scope of that manager).
        #     Must not be `None`.
        #   uri (string):
        #     Canonical URI path of the resource.
        #     Must not be `None`.
        #   name (string):
        #     Name of the resource.
        #     May be `None`.
        #   properties (dict):
        #     Properties for this resource object. May be `None` or empty.
        #     * Key: Name of the property.
        #     * Value: Value of the property.

        # We want to surface precondition violations as early as possible,
        # so we test those that are not surfaced through the init code:
        assert manager is not None
        assert uri is not None

        self._manager = manager
        self._uri = uri

        self._properties = dict(properties) if properties else {}
        if name is not None:
            name_prop = self._manager._name_prop
            if name_prop in self._properties:
                assert self._properties[name_prop] == name
            else:
                self._properties[name_prop] = name
        uri_prop = self._manager._uri_prop
        if uri_prop in self._properties:
            assert self._properties[uri_prop] == uri
        else:
            self._properties[uri_prop] = uri

        self._properties_timestamp = int(time.time())
        self._full_properties = False
        # self._property_lock = contextlib.nullcontext()  # test need to lock
        self._property_lock = threading.RLock()
        self._auto_update = False
        self._ceased_existence = False

    @property
    def properties(self):
        """
        :class:`iv:immutable_views.DictView`: The properties of this resource
        that are currently present in this Python object, as a dictionary.

          * Key: Name of the property.
          * Value: Value of the property.

        The returned :class:`iv:immutable_views.DictView` object is an immutable
        dictionary view that behaves like a standard Python :class:`dict`
        except that it prevents any modifications to the dictionary.

        See the respective 'Data model' sections in the :term:`HMC API` book
        for a description of the resources along with their properties.

        The dictionary contains either the full set of resource properties,
        or a subset thereof, or can be empty in some cases.

        Because the presence of properties in this dictionary depends on the
        situation, the purpose of this dictionary is only for iterating
        through the resource properties that are currently present.

        Specific resource properties should be accessed via:

        * The resource name, via the :attr:`~zhmcclient.BaseResource.name`
          attribute.
        * The resource URI, via the :attr:`~zhmcclient.BaseResource.uri`
          attribute.
        * Any resource property, via the
          :meth:`~zhmcclient.BaseResource.get_property` or
          :meth:`~zhmcclient.BaseResource.prop` methods.

        Updates to property values can be done via the ``update_properties()``
        method of the resource class. Which properties can be updated
        is indicated with the 'w' qualifier in the data model of the resource
        in the :term:`HMC API` book.

        If :ref:`auto-updating` is enabled for the
        resource object and the session is enabled for auto-updating as well,
        the property values in the returned :class:`iv:immutable_views.DictView`
        object will change as they change on the HMC.

        If the resource object on the HMC no longer exists, the properties
        show the values that were last updated from the HMC when the object
        still existed.
        """
        return DictView(self._properties)

    @property
    def uri(self):
        """
        string: The canonical URI path of the resource. Will not be `None`.

        Example: ``/api/cpcs/12345``
        """
        return self._uri

    @property
    def name(self):
        """
        string: The name of the resource. Will not be `None`.

        The resource name is unique across its sibling resources of the same
        type and with the same parent resource.

        Accessing this property will cause the properties of this resource
        object to be updated from the HMC, if it does not yet contain the
        property for the resource name.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.CeasedExistence`: Only when not yet available
            locally.
        """
        # pylint: disable=protected-access
        # We avoid storing the name in an instance variable, because it can
        # be modified via update_properties().
        name_prop = self.manager._name_prop
        try:
            with self._property_lock:
                return self._properties[name_prop]
        except KeyError:
            self.pull_full_properties()
            with self._property_lock:
                return self._properties[name_prop]

    @property
    def manager(self):
        """
        Subclass of :class:`~zhmcclient.BaseManager`:
          Manager object for this resource (and for all resources of the same
          type in the scope of that manager). Will not be `None`.
        """
        return self._manager

    @property
    def full_properties(self):
        """
        A boolean indicating whether or not the resource properties in this
        object are the full set of resource properties.

        Note that listing resources and creating new resources produces objects
        that have less than the full set of properties.
        """
        return self._full_properties

    @property
    def properties_timestamp(self):
        """
        The point in time of the last update of the resource properties cached
        in this object, as Unix time (an integer that is the number of seconds
        since the Unix epoch).
        """
        return self._properties_timestamp

    @property
    def ceased_existence(self):
        """
        bool: Indicates that the corresponding object on the HMC no longer
        exists, if auto-update is enabled for the resource. Always `False`, if
        auto-update is not enabled for the resource.
        """
        return self._ceased_existence

    @logged_api_call
    def pull_full_properties(self):
        """
        Retrieve the full set of resource properties from the HMC and cache
        them in this Python object.

        If the resource no longer exists on the HMC,
        :exc:`~zhmcclient.CeasedExistence` will be raised.

        This method serializes with other methods that access or change
        resource properties on the same Python object.

        Authorization requirements:

        * Object-access permission to this resource.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.CeasedExistence`
        """
        with self._property_lock:
            if self._ceased_existence:
                raise CeasedExistence(self._uri)

        try:
            full_properties = self.manager.session.get(self._uri, resource=self)
        except HTTPError as exc:
            if exc.http_status == 404 and exc.reason == 1:
                # The resource no longer exists
                self.cease_existence_local()
                raise CeasedExistence(self._uri)
            raise

        with self._property_lock:
            self._properties.update(full_properties)
            self._properties_timestamp = int(time.time())
            self._full_properties = True

    @logged_api_call
    def pull_properties(self, properties):
        """
        Retrieve the specified set of resource properties from the HMC and
        cache them in this Python object.

        If no properties are specified, the method does nothing.

        The values of other properties that may already be cached in this
        Python object remain unchanged.

        If the HMC does not support property filtering for this type of
        resource, or if the resource on the HMC does not have one or more of
        the specified properties, the full set of properties is retrieved from
        the HMC and cached in this Python object.

        If the resource no longer exists on the HMC,
        :exc:`~zhmcclient.CeasedExistence` will be raised.

        This method serializes with other methods that access or change
        resource properties on the same Python object.

        Authorization requirements:

        * Object-access permission to this resource.

        Parameters:

          properties (:term:`string` or list/tuple of strings):
            Single name or list/tuple of names of the resource properties, using
            the names defined in the respective 'Data model' sections in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.CeasedExistence`
        """
        if not properties:
            return

        if not isinstance(properties, (list, tuple)):
            properties = [properties]

        with self._property_lock:
            if self._ceased_existence:
                raise CeasedExistence(self._uri)

        if self.manager.supports_properties:
            # Note: Older HMC versions may simply ignore the query parameter
            # and return the full list of properties. Newer HMC versions return
            # HTTP 400,1 "unrecognized or unsupported query parameter" if it is
            # not supported for the resource type.
            uri = f"{self._uri}?properties={','.join(properties)}"
            try:
                subset_properties = self.manager.session.get(uri, resource=self)
                # pylint: disable=simplifiable-if-statement
                if len(subset_properties) > len(properties):
                    # We have an older HMC that ignored the query parameter and
                    # returned all properties.
                    is_full = True
                else:
                    is_full = False
            except HTTPError as exc:
                if exc.http_status == 400 and exc.reason == 1:
                    # HMC does not yet support the query parameter, get full set
                    subset_properties = self.manager.session.get(
                        self._uri, resource=self)
                    is_full = True
                elif exc.http_status == 400 and exc.reason == 14:
                    # The resource does not have one or more of the specified
                    # properties, get full set
                    subset_properties = self.manager.session.get(
                        self._uri, resource=self)
                    is_full = True
                elif exc.http_status == 404 and exc.reason == 1:
                    # The resource no longer exists
                    self.cease_existence_local()
                    raise CeasedExistence(self._uri)
                else:
                    raise
        else:
            # Resource does not support the query parameter, get full set
            try:
                subset_properties = self.manager.session.get(
                    self._uri, resource=self)
                is_full = True
            except HTTPError as exc:
                if exc.http_status == 404 and exc.reason == 1:
                    # The resource no longer exists
                    self.cease_existence_local()
                    raise CeasedExistence(self._uri)
                raise

        with self._property_lock:
            if is_full:
                self._properties.update(subset_properties)
                self._properties_timestamp = int(time.time())
                self._full_properties = True
            else:
                self._properties.update(subset_properties)
                # We leave the self._full_properties flag unchanged. If the
                # local object already had full properties pulled earlier, it
                # now still has all of them.
                # We leave the self._properties_timestamp unchanged. The
                # resource now has newer and older properties, and the timestamp
                # indicates the oldest properties.

    @logged_api_call
    def get_property(self, name):
        """
        Return the value of a resource property.

        If the resource property is not cached in this Python object, the full
        set of resource properties is retrieved from the HMC and cached in this
        Python object.

        If the resource no longer exists on the HMC,
        :exc:`~zhmcclient.CeasedExistence` will be raised.

        If the resource on the HMC does not have the property, :exc:`KeyError`
        will be raised.

        This method serializes with other methods that access or change
        resource properties on the same Python object.

        Authorization requirements:

        * Object-access permission to this resource.

        Parameters:

          name (:term:`string`):
            Name of the resource property, using the names defined in the
            respective 'Data model' sections in the :term:`HMC API` book.

        Returns:

          The value of the resource property.

        Raises:

          KeyError: The resource does not have a property with that name.
          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.CeasedExistence`
        """
        try:
            with self._property_lock:
                if self._ceased_existence:
                    raise CeasedExistence(self._uri)
                return self._properties[name]
        except KeyError:
            if self._full_properties:
                raise
            self.pull_full_properties()
            with self._property_lock:
                return self._properties[name]

    @logged_api_call
    def prop(self, name, default=None):
        """
        Return the value of a resource property, applying a default if the
        resource does not have a property with that name.

        If the resource property is cached in this Python object, its value is
        used. Otherwise, the full set of resource properties is retrieved from
        the HMC and cached in this Python object.

        If the resource no longer exists on the HMC,
        :exc:`~zhmcclient.CeasedExistence` is raised.

        This method serializes with other methods that access or change
        resource properties on the same Python object.

        Authorization requirements:

        * Object-access permission to this resource.

        Parameters:

          name (:term:`string`):
            Name of the resource property, using the names defined in the
            respective 'Data model' sections in the :term:`HMC API` book.

          default:
            Default value to be used, if the resource does not have a property
            with the specified name.

        Returns:

          The value of the resource property.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.CeasedExistence`
        """
        try:
            return self.get_property(name)
        except KeyError:
            return default

    def get_properties_pulled(self, names):
        """
        Return a set of resource properties with the values they currently have
        on the HMC.

        If auto-update is enabled for the resource, the property value is
        returned from the (automatically updated) local cache. Otherwise, a
        "Get Properties" operation is performed to get the current values. If
        supported by the type of resource, the operation uses the 'properties'
        query parameter to restrict the response to the desired properties.

        If the resource on the HMC does not have a property that was specified,
        its value will default to `None`.

        If the resource no longer exists on the HMC,
        :exc:`~zhmcclient.CeasedExistence` is raised.

        This method serializes with other methods that access or change
        resource properties on the same Python object.

        Authorization requirements:

        * Object-access permission to this resource.

        Parameters:

          names (:term:`string` or list/tuple of strings):
            Single name or list/tuple of names of the resource properties, using
            the names defined in the respective 'Data model' sections in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.CeasedExistence`
        """
        if not self.auto_update_enabled():
            self.pull_properties(names)

        return self.get_properties_local(names)

    def get_properties_local(self, names, defaults=None):
        """
        Return the values of a set of resource properties, using default values
        for those that are not cached in this Python object, without
        retrieving them from the HMC.

        This method serializes with other methods that access or change
        resource properties on the same Python object.

        Parameters:

          names (:term:`string` or list/tuple of strings):
            Single name or list/tuple of names of the resource properties, using
            the names defined in the respective 'Data model' sections in the
            :term:`HMC API` book.

          defaults:
            Single value or list/tuple of values to be used as a default for
            resource properties that are not cached in this Python object. If a
            single value, it is used for all properties that are not cached.
            If a list/tuple, it must be index-correlated with the names
            list/tuple.

        Returns:

          Single value (if names is a single value) or list of values (if names
          is a list/tuple of values) of the properties, with defaults applied.
        """
        if isinstance(names, (list, tuple)):
            with self._property_lock:
                values = []
                for i, name in enumerate(names):
                    try:
                        value = self._properties[name]
                    except KeyError:
                        if isinstance(defaults, (list, tuple)):
                            value = defaults[i]
                        else:
                            value = defaults
                    values.append(value)
                return values
        else:
            with self._property_lock:
                try:
                    return self._properties[names]
                except KeyError:
                    return defaults

    def update_properties_local(self, properties):
        """
        Update the values of a set of resource properties on this Python object
        without propagating the updates to the HMC.

        If a property to be updated is not present in the Python object, it
        is added.

        This method serializes with other methods that access or change
        resource properties on the same Python object.

        Parameters:

          properties (:class:`py:dict`):
            Dictionary of new property values, with:

            - Key: Name of the property, using the names defined in the
              respective 'Data model' sections in the :term:`HMC API` book.

            - Value: New value for the property.
        """
        with self._property_lock:
            for name, value in properties.items():
                self._properties[name] = value

    def cease_existence_local(self):
        """
        Update this Python object to indicate that the corresponding HMC object
        no longer exists.

        This method serializes with other methods that access or change
        resource properties on the same Python object.
        """
        with self._property_lock:
            self._ceased_existence = True

    def __str__(self):
        """
        Return a human readable string representation of this resource.

        This method serializes with other methods that access or change
        resource properties on the same Python object.

        Example result:

        .. code-block:: text

            Cpc(name=P0000S12,
                object-uri=/api/cpcs/f1bc49af-f71a-3467-8def-3c186b5d9352,
                status=service-required)
        """
        with self._property_lock:
            properties_keys = self._properties.keys()
            search_keys = ['status', 'object-uri', 'element-uri', 'name',
                           'type', 'class']
            sorted_keys = sorted([k for k in properties_keys
                                  if k in search_keys])
            info = ", ".join(f"{k}={self._properties[k]!r}"
                             for k in sorted_keys)
            return f"{self.__class__.__name__}({info})"

    def __repr__(self):
        """
        Return a string with the state of this resource, for debug purposes.

        This method serializes with other methods that access or change
        resource properties on the same Python object.

        Note that the derived resource classes that have child resources
        have their own ``__repr__()`` methods, because only they know which
        child resources they have.
        """
        with self._property_lock:
            ret = (
                f"{repr_obj_id(self)} (\n"
                f"  _manager={repr_obj_id(self._manager)},\n"
                f"  _uri={self._uri!r},\n"
                f"  _ceased_existence={self._ceased_existence!r},\n"
                f"  _full_properties={self._full_properties!r},\n"
                f"  _auto_update={self._auto_update!r},\n"
                f"  _properties_timestamp="
                f"{repr_timestamp(self._properties_timestamp)},\n"
                f"  _properties={repr_dict(self._properties, indent=4)}\n"
                ")")
            return ret

    def auto_update_enabled(self):
        """
        Return whether :ref:`auto-updating` is
        currently enabled for the resource object.

        Return:
          bool: Indicates whether auto-update is enabled.
        """
        return self._auto_update

    def enable_auto_update(self):
        """
        Enable :ref:`auto-updating` for this
        resource object, if currently disabled.

        When enabling auto-update, the session to which this resource belongs is
        subscribed for auto-updating if needed (see
        :meth:`~zhmcclient.Session.subscribe_auto_update`), the resource
        object is registered with the session's auto updater via
        :meth:`~zhmcclient.AutoUpdater.register_object`, and all properties
        of this resource object are retrieved from the HMC using
        :meth:`pull_full_properties` in order to have the most current values
        as a basis for the future auto-updating.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.CeasedExistence`
        """
        if not self._auto_update:
            session = self.manager.session
            session.subscribe_auto_update()
            session.auto_updater.register_object(self)
            self._auto_update = True
            self.pull_full_properties()

    def disable_auto_update(self):
        """
        Disable :ref:`auto-updating` for this
        resource object, if currently enabled.

        When disabling auto-updating, the resource object is unregistered from
        the session's auto updater via
        :meth:`~zhmcclient.AutoUpdater.unregister_object`, and the session
        is unsubscribed from auto-updating if the auto updater has no more
        objects registered.
        """
        if self._auto_update:
            self._auto_update = False
            session = self.manager.session
            session.auto_updater.unregister_object(self)
            if not session.auto_updater.has_objects():
                session.unsubscribe_auto_update()

    def dump(self):
        """
        Dump this resource with its properties and child resources
        (recursively) as a resource definition.

        This is the default implementation for the case where the resource has
        no child resources. If the resource does have child resources, this
        method needs to be overridden in the resource subclass.

        The returned resource definition of this implementation has the
        following format::

            {
                "properties": {...},
            }

        Returns:

          dict: Resource definition of this resource.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
          :exc:`~zhmcclient.CeasedExistence`
        """
        resource_dict = OrderedDict()
        self.pull_full_properties()
        resource_dict['properties'] = OrderedDict(self._properties)
        # No child resources
        return resource_dict



# Copyright 2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _auto_updater module.
"""


import time
import re
import logging
from unittest.mock import patch
import pytest

from zhmcclient import Client
from zhmcclient_mock import FakedSession

from .test_notification import MockedStompConnection

# Test resources that are the basis, in FakedHMC.add_resources() format
TEST_RESOURCES_BASE = {
    'cpcs': [
        {
            'properties': {
                'object-uri': '/api/cpcs/fake-cpc1',
                'object-id': 'fake-cpc1',
                'name': 'fake-cpc1',  # pc (but ro)
                'dpm-enabled': True,  # pc
                'description': 'description1',  # pc if DPM
                'acceptable-status': ['active'],  # pc
                'degraded-status': [],  # pc
                'status': 'active',  # sc
                'additional-status': 'addstatus',  # sc (not in data model)
                'has-unacceptable-status': False,  # sc
            },
            'partitions': [
                {
                    'properties': {
                        'object-uri': '/api/partitions/fake-part1',
                        'object-id': 'fake-part1',
                        'name': 'fake-part1',  # pc
                        'description': 'description1',  # pc
                        'acceptable-status': ['active'],  # pc
                        'status': 'active',  # sc
                        'additional-status': 'addstatus',  # sc (not in model)
                        'has-unacceptable-status': False,  # sc
                    },
                    'nics': [
                        {
                            'properties': {
                                'element-uri':
                                '/api/partitions/fake-part1/nics/fake-nic1',
                                'element-id': 'fake-nic1',
                                'name': 'fake-nic1',  # pc
                                'description': 'description1',  # pc
                            },
                        },
                    ],
                },
            ],
        },
    ],
    'consoles': [
        {
            'properties': {
                'object-uri': '/api/console',
                'name': 'fake-hmc1',
                'description': 'fake-hmc1',
                'version': '2.15.0',
            },
            'unmanaged_cpcs': [],
        },
    ],
}

# Test partition 2 that gets added, in FakedPartitionManager.add() format
TEST_PARTITION_2 = {
    'object-uri': '/api/partitions/fake-part2',
    'object-id': 'fake-part2',
    'name': 'fake-part2',  # pc
    'description': 'description2',  # pc
    'acceptable-status': ['active'],  # pc
    'status': 'active',  # sc
    'additional-status': 'addstatus',  # sc (not in model)
    'has-unacceptable-status': False,  # sc
}

# Test CPC 2 that gets added, in FakedCpcManager.add() format
TEST_CPC_2 = {
    'object-uri': '/api/cpcs/fake-cpc2',
    'object-id': 'fake-cpc2',
    'name': 'fake-cpc2',  # pc (but ro)
    'dpm-enabled': True,  # pc
    'description': 'description2',  # pc if DPM
    'acceptable-status': ['active'],  # pc
    'degraded-status': [],  # pc
    'status': 'inactive',  # sc
    'additional-status': 'addstatus',  # sc (not in data model)
    'has-unacceptable-status': True,  # sc
}


# Testcases for test_auto_updater_all()
TESTCASES_AUTO_UPDATER_ALL = [
    # Each list item is a separate testcase for which test_auto_updater_all()
    # is called.
    # A testcase item is a tuple with these members:
    # * string: testcase description
    # * dict: testcase data, with the following dict items:
    #   - initial_resources: Initial mock resources to be set up before the
    #     test, in FakedBaseResource.add_resources() format.
    #   - auto_updated_uris: List of canonical resource or manager URIs to
    #     enable for auto-updating before the test. The format for manager
    #     URIs must follow the implementation in BaseManager.uri.
    #   - added_resources: List of tuple(parent_uri, resources) for
    #     resources to be added to trigger inventory add notifications.
    #     child_resources is in FakedBaseResource.add_resources() format.
    #   - removed_resources: List of resource_uri for resources to be removed
    #     to trigger inventory remove notifications.
    #   - updated_resources: List of tuple(resource_uri, properties) for
    #     resources to be updated to trigger property change or status change
    #     notifications.
    #   - notifications: List of tuple(header, message) for JMS notifications
    #     that are triggered. header and message are Python objects that can be
    #     dumped to JSON.
    #   - exp_lists: List of tuple(mgr_uri, res_uris, local_res_uris) for the
    #     expected resources and local resources in a manager after the changes.
    #   - exp_resources: List of tuple(uri, properties) for expected resource
    #     properties after the changes.
    #   - exp_log_entries: List of tuple(logger_name, level, message_pattern)
    #     for expected log entries. Only these are verified, any additional
    #     log entries are tolerated.

    (
        "no auto-updated resources, no changes, no notifications",
        dict(
            initial_resources=TEST_RESOURCES_BASE,
            auto_updated_uris=[],
            added_resources=[],
            removed_resources=[],
            updated_resources=[],
            notifications=[],
            exp_lists=[],
            exp_resources=[],
            exp_log_entries=[],
        ),
    ),
    (
        "property change notification for an auto-updated CPC "
        "(top-level object)",
        dict(
            initial_resources=TEST_RESOURCES_BASE,
            auto_updated_uris=[
                '/api/cpcs/fake-cpc1',
            ],
            added_resources=[],
            removed_resources=[],
            updated_resources=[
                (
                    '/api/cpcs/fake-cpc1',
                    {
                        'description': 'desc_new',
                    },
                ),
            ],
            notifications=[
                (
                    {
                        'notification-type': 'property-change',
                        'object-uri': '/api/cpcs/fake-cpc1',
                    },
                    {
                        'change-reports': [
                            {
                                'property-name': 'description',
                                'old-value': 'description1',
                                'new-value': 'desc_new',
                            },
                        ],
                    },
                ),
            ],
            exp_lists=[
                (
                    '/#cpc',
                    [
                        '/api/cpcs/fake-cpc1',
                    ],
                    [
                        '/api/cpcs/fake-cpc1',
                    ],
                ),
            ],
            exp_resources=[
                (
                    '/api/cpcs/fake-cpc1',
                    {
                        'description': 'desc_new',
                    },
                ),
            ],
            exp_log_entries=[
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* established"),
                ('zhmcclient.jms', logging.DEBUG,
                 "JMS message for property change notification .*"
                 "resource /api/cpcs/fake-cpc1 .*"
                 "u?'property-name': u?'description'"),
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* disconnected"),
            ],
        ),
    ),
    (
        "property change notification for a non auto-updated CPC",
        dict(
            initial_resources=TEST_RESOURCES_BASE,
            auto_updated_uris=[],  # Not auto-updated
            added_resources=[],
            removed_resources=[],
            updated_resources=[
                (
                    '/api/cpcs/fake-cpc1',
                    {
                        'description': 'description1',  # Old value
                    },
                ),
            ],
            notifications=[
                (
                    {
                        'notification-type': 'property-change',
                        'object-uri': '/api/cpcs/fake-cpc1',
                    },
                    {
                        'change-reports': [
                            {
                                'property-name': 'description',
                                'old-value': 'description1',
                                'new-value': 'desc_new',
                            },
                        ],
                    },
                ),
            ],
            exp_lists=[
                (
                    '/#cpc',
                    [
                        '/api/cpcs/fake-cpc1',
                    ],
                    [
                        '/api/cpcs/fake-cpc1',
                    ],
                ),
            ],
            exp_resources=[
                (
                    '/api/cpcs/fake-cpc1',
                    {
                        'description': 'description1',
                    },
                ),
            ],
            exp_log_entries=[],
        ),
    ),
    (
        "property change notification with two changes",
        dict(
            initial_resources=TEST_RESOURCES_BASE,
            auto_updated_uris=[
                '/api/cpcs/fake-cpc1',
            ],
            added_resources=[],
            removed_resources=[],
            updated_resources=[
                (
                    '/api/cpcs/fake-cpc1',
                    {
                        'description': 'desc_new',
                        'degraded-status': ['memory'],
                    },
                ),
            ],
            notifications=[
                (
                    {
                        'notification-type': 'property-change',
                        'object-uri': '/api/cpcs/fake-cpc1',
                    },
                    {
                        'change-reports': [
                            {
                                'property-name': 'description',
                                'old-value': 'description1',
                                'new-value': 'desc_new',
                            },
                            {
                                'property-name': 'degraded-status',
                                'old-value': [],
                                'new-value': ['memory'],
                            },
                        ],
                    },
                ),
            ],
            exp_lists=[
                (
                    '/#cpc',
                    [
                        '/api/cpcs/fake-cpc1',
                    ],
                    [
                        '/api/cpcs/fake-cpc1',
                    ],
                ),
            ],
            exp_resources=[
                (
                    '/api/cpcs/fake-cpc1',
                    {
                        'description': 'desc_new',
                        'degraded-status': ['memory'],
                    },
                ),
            ],
            exp_log_entries=[
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* established"),
                ('zhmcclient.jms', logging.DEBUG,
                 "JMS message for property change notification .*"
                 "resource /api/cpcs/fake-cpc1 .*"
                 "u?'property-name': u?'description'.*"
                 "u?'property-name': u?'degraded-status'"),
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* disconnected"),
            ],
        ),
    ),
    (
        "property change notification on an auto-updated NIC "
        "(element resource)",
        dict(
            initial_resources=TEST_RESOURCES_BASE,
            auto_updated_uris=[
                '/api/partitions/fake-part1/nics/fake-nic1',
            ],
            added_resources=[],
            removed_resources=[],
            updated_resources=[
                (
                    '/api/partitions/fake-part1/nics/fake-nic1',
                    {
                        'description': 'desc_new',
                    },
                ),
            ],
            notifications=[
                (
                    {
                        'notification-type': 'property-change',
                        'element-uri':
                        '/api/partitions/fake-part1/nics/fake-nic1',
                    },
                    {
                        'change-reports': [
                            {
                                'property-name': 'description',
                                'old-value': 'description1',
                                'new-value': 'desc_new',
                            },
                        ]
                    },
                ),
            ],
            exp_lists=[
                (
                    '/api/partitions/fake-part1#nic',
                    [
                        '/api/partitions/fake-part1/nics/fake-nic1',
                    ],
                    [
                        '/api/partitions/fake-part1/nics/fake-nic1',
                    ],
                ),
            ],
            exp_resources=[
                (
                    '/api/partitions/fake-part1/nics/fake-nic1',
                    {
                        'description': 'desc_new',
                    },
                ),
            ],
            exp_log_entries=[
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* established"),
                ('zhmcclient.jms', logging.DEBUG,
                 "JMS message for property change notification .*"
                 "resource /api/partitions/fake-part1/nics/fake-nic1 .*"
                 "u?'property-name': u?'description'"),
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* disconnected"),
            ],
        ),
    ),
    (
        "status change notification on an auto-updated CPC "
        "(top-level resource)",
        dict(
            initial_resources=TEST_RESOURCES_BASE,
            auto_updated_uris=[
                '/api/cpcs/fake-cpc1',
            ],
            added_resources=[],
            removed_resources=[],
            updated_resources=[
                (
                    '/api/cpcs/fake-cpc1',
                    {
                        'status': 'inactive',
                        'additional-status': 'addstatus_new',
                        'has-unacceptable-status': True,
                    },
                ),
            ],
            notifications=[
                (
                    {
                        'notification-type': 'status-change',
                        'object-uri': '/api/cpcs/fake-cpc1',
                    },
                    {
                        'change-reports': [
                            {
                                'old-status': 'active',
                                'old-additional-status': 'addstatus',
                                'new-status': 'inactive',
                                'new-additional-status': 'addstatus_new',
                                'has-unacceptable-status': True,
                            },
                        ]
                    },
                ),
            ],
            exp_lists=[
                (
                    '/#cpc',
                    [
                        '/api/cpcs/fake-cpc1',
                    ],
                    [
                        '/api/cpcs/fake-cpc1',
                    ],
                ),
            ],
            exp_resources=[
                (
                    '/api/cpcs/fake-cpc1',
                    {
                        'status': 'inactive',
                        'additional-status': 'addstatus_new',
                        'has-unacceptable-status': True,
                    },
                ),
            ],
            exp_log_entries=[
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* established"),
                ('zhmcclient.jms', logging.DEBUG,
                 "JMS message for status change notification"),
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* disconnected"),
            ],
        ),
    ),
    (
        "inventory change notification for removing a partition from an "
        "auto-updated partition manager (object resource)",
        dict(
            initial_resources=TEST_RESOURCES_BASE,
            auto_updated_uris=[
                '/api/cpcs/fake-cpc1#partition',
            ],
            added_resources=[],
            removed_resources=[
                (
                    '/api/cpcs/fake-cpc1',
                    'partitions',
                    'fake-part1',
                ),
            ],
            updated_resources=[],
            notifications=[
                (
                    {
                        'notification-type': 'inventory-change',
                        'object-uri': '/api/partitions/fake-part1',
                        'object-id': 'fake-part1',
                        'class': 'partition',
                        'action': 'remove',
                    },
                    None,
                ),
            ],
            exp_lists=[
                (
                    '/api/cpcs/fake-cpc1#partition',
                    [],
                    [],
                ),
            ],
            exp_resources=[],
            exp_log_entries=[
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* established"),
                ('zhmcclient.jms', logging.DEBUG,
                 "JMS message for inventory change notification .*"
                 "resource /api/partitions/fake-part1 .*"
                 "action: u?'remove'"),
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* disconnected"),
            ],
        ),
    ),
    (
        "inventory change notification for adding a partition to an "
        "auto-updated partition manager (object resource)",
        dict(
            initial_resources=TEST_RESOURCES_BASE,
            auto_updated_uris=[
                '/api/cpcs/fake-cpc1#partition',
            ],
            added_resources=[
                (
                    '/api/cpcs/fake-cpc1',
                    {
                        'partitions': [
                            {
                                'properties': TEST_PARTITION_2,
                            },
                        ],
                    },
                ),
            ],
            removed_resources=[],
            updated_resources=[],
            notifications=[
                (
                    {
                        'notification-type': 'inventory-change',
                        'object-uri': TEST_PARTITION_2['object-uri'],
                        'object-id': TEST_PARTITION_2['object-id'],
                        'class': 'partition',
                        'action': 'add',
                    },
                    None,
                ),
            ],
            exp_lists=[
                (
                    '/api/cpcs/fake-cpc1#partition',
                    [
                        '/api/partitions/fake-part1',
                        TEST_PARTITION_2['object-uri'],
                    ],
                    [
                        '/api/partitions/fake-part1',
                        TEST_PARTITION_2['object-uri'],
                    ],
                ),
            ],
            exp_resources=[
                (
                    TEST_PARTITION_2['object-uri'],
                    TEST_PARTITION_2,
                ),
            ],
            exp_log_entries=[
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* established"),
                ('zhmcclient.jms', logging.DEBUG,
                 "JMS message for inventory change notification .*"
                 "resource /api/partitions/fake-part2 .*"
                 "action: u?'add'"),
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* disconnected"),
            ],
        ),
    ),
    (
        "inventory change notification for removing a CPC from an "
        "auto-updated CPC manager (top-level resource)",
        dict(
            initial_resources=TEST_RESOURCES_BASE,
            auto_updated_uris=[
                '/#cpc',
            ],
            added_resources=[],
            removed_resources=[
                (
                    None,
                    'cpcs',
                    'fake-cpc1',
                ),
            ],
            updated_resources=[],
            notifications=[
                (
                    {
                        'notification-type': 'inventory-change',
                        'object-uri': '/api/cpcs/fake-cpc1',
                        'object-id': 'fake-cpc1',
                        'class': 'cpc',
                        'action': 'remove',
                    },
                    None,
                ),
            ],
            exp_lists=[
                (
                    '/#cpc',
                    [],
                    [],
                ),
            ],
            exp_resources=[],
            exp_log_entries=[
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* established"),
                ('zhmcclient.jms', logging.DEBUG,
                 "JMS message for inventory change notification .*"
                 "resource /api/cpcs/fake-cpc1 .*"
                 "action: u?'remove'"),
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* disconnected"),
            ],
        ),
    ),
    (
        "inventory change notification for adding a CPC to an "
        "auto-updated CPC manager ((top-level resource)",
        dict(
            initial_resources=TEST_RESOURCES_BASE,
            auto_updated_uris=[
                '/#cpc',
            ],
            added_resources=[
                (
                    None,
                    {
                        'cpcs': [
                            {
                                'properties': TEST_CPC_2,
                            },
                        ],
                    },
                ),
            ],
            removed_resources=[],
            updated_resources=[],
            notifications=[
                (
                    {
                        'notification-type': 'inventory-change',
                        'object-uri': TEST_CPC_2['object-uri'],
                        'object-id': TEST_CPC_2['object-id'],
                        'class': 'cpc',
                        'action': 'add',
                    },
                    None,
                ),
            ],
            exp_lists=[
                (
                    '/#cpc',
                    [
                        '/api/cpcs/fake-cpc1',
                        TEST_CPC_2['object-uri'],
                    ],
                    [
                        '/api/cpcs/fake-cpc1',
                        TEST_CPC_2['object-uri'],
                    ],
                ),
            ],
            exp_resources=[
                (
                    TEST_CPC_2['object-uri'],
                    TEST_CPC_2,
                ),
            ],
            exp_log_entries=[
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* established"),
                ('zhmcclient.jms', logging.DEBUG,
                 "JMS message for inventory change notification .*"
                 "resource /api/cpcs/fake-cpc2 .*"
                 "action: u?'add'"),
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* disconnected"),
            ],
        ),
    ),
    (
        "inventory change notification with invalid action "
        "(that would be an HMC error or a future extension)",
        dict(
            initial_resources=TEST_RESOURCES_BASE,
            auto_updated_uris=[
                '/api/cpcs/fake-cpc1#partition',
            ],
            added_resources=[],
            removed_resources=[],
            updated_resources=[],
            notifications=[
                (
                    {
                        'notification-type': 'inventory-change',
                        'object-uri': '/api/partitions/fake-part1',
                        'object-id': 'fake-part1',
                        'class': 'partition',
                        'action': 'xyz',  # invalid action
                    },
                    None,
                ),
            ],
            exp_lists=[
                (
                    '/api/cpcs/fake-cpc1#partition',
                    [
                        '/api/partitions/fake-part1',
                    ],
                    [
                        '/api/partitions/fake-part1',
                    ],
                ),
            ],
            exp_resources=[],
            exp_log_entries=[
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* established"),
                ('zhmcclient.jms', logging.DEBUG,
                 "JMS message for inventory change notification .*"
                 "resource /api/partitions/fake-part1 .*"
                 "action: u?'xyz'"),
                ('zhmcclient.jms', logging.ERROR,
                 "JMS message for inventory change notification .*"
                 "unknown action 'xyz'"),
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* disconnected"),
            ],
        ),
    ),
    (
        "property change notification that misses both object-uri and "
        "element-uri (that would be an HMC error)",
        dict(
            initial_resources=TEST_RESOURCES_BASE,
            auto_updated_uris=[
                '/api/cpcs/fake-cpc1',
            ],
            added_resources=[],
            removed_resources=[],
            updated_resources=[],
            notifications=[
                (
                    {
                        'notification-type': 'property-change',
                        # missing element-uri and object-uri
                    },
                    {
                        'change-reports': [
                            {
                                'property-name': 'description',
                                'old-value': 'description1',
                                'new-value': 'desc_new',
                            },
                        ]
                    },
                ),
            ],
            exp_lists=[
                (
                    '/#cpc',
                    [
                        '/api/cpcs/fake-cpc1',
                    ],
                    [
                        '/api/cpcs/fake-cpc1',
                    ],
                ),
            ],
            exp_resources=[
                (
                    '/api/cpcs/fake-cpc1',
                    {
                        'description': 'description1',
                    },
                ),
            ],
            exp_log_entries=[
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* established"),
                ('zhmcclient.jms', logging.ERROR,
                 "JMS message for object notification .*"
                 "no 'element-uri' .*no 'object-uri'"),
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* disconnected"),
            ],
        ),
    ),
    (
        "valid notification unrelated to auto-updating (should be ignored)",
        dict(
            initial_resources=TEST_RESOURCES_BASE,
            auto_updated_uris=[
                '/api/cpcs/fake-cpc1',
            ],
            added_resources=[],
            removed_resources=[],
            updated_resources=[],
            notifications=[
                (
                    {
                        'notification-type': 'job-completion',  # unrelated
                        'job-uri': 'bla',
                    },
                    None,
                ),
            ],
            exp_lists=[],
            exp_resources=[],
            exp_log_entries=[
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* established"),
                ('zhmcclient.jms', logging.WARNING,
                 "JMS message for notification of type job-completion .*"
                 "is ignored"),
                ('zhmcclient.jms', logging.INFO,
                 "JMS session .* disconnected"),
            ],
        ),
    ),
]


@pytest.mark.parametrize(
    "desc, testcase",
    TESTCASES_AUTO_UPDATER_ALL)
@patch(target='stomp.Connection', new=MockedStompConnection)
def test_auto_updater_all(desc, testcase, caplog):
    # pylint: disable=unused-argument
    """
    Test function for an overall use of AutoUpdater with auto-updated
    resources.
    """

    # In this function, the variables for faked zhmcclient objects are named
    # 'faked_*'.

    initial_resources = testcase['initial_resources']
    auto_updated_uris = testcase['auto_updated_uris']
    added_resources = testcase['added_resources']
    removed_resources = testcase['removed_resources']
    updated_resources = testcase['updated_resources']
    notifications = testcase['notifications']
    exp_lists = testcase['exp_lists']
    exp_resources = testcase['exp_resources']
    exp_log_entries = testcase['exp_log_entries']

    caplog.set_level(logging.DEBUG, logger="zhmcclient.jms")

    # Create a zhmcclient mock environment
    faked_session = FakedSession(
        'fake-hmc', 'fake-hmc', 'fake-version', 'fake-version')
    client = Client(faked_session)

    # Add the initial resources for the testcase to the mock enviroment.
    # They will be retrieved in order to enable auto-updating for them.
    faked_session.hmc.add_resources(initial_resources)

    # List the resources and managerss
    managers = {}  # key: manager URI, value: zhmcclient manager object
    resource_objs = []  # zhmcclient resource objects
    cpcs = client.cpcs.list()
    resource_objs.extend(cpcs)
    managers[client.cpcs.uri] = client.cpcs
    for cpc in cpcs:
        partitions = cpc.partitions.list()
        resource_objs.extend(partitions)
        managers[cpc.partitions.uri] = cpc.partitions
        for partition in partitions:
            nics = partition.nics.list()
            resource_objs.extend(nics)
            managers[partition.nics.uri] = partition.nics

    # Enable the desired resource and manager objects for auto-updating.
    # This causes them to be registered with the auto updater.
    enabled_res_objs = []
    enabled_mgr_objs = []
    for res_obj in resource_objs:
        if res_obj.uri in auto_updated_uris:
            enabled_res_objs.append(res_obj)
    for mgr_uri, mgr_obj in managers.items():
        if mgr_uri in auto_updated_uris:
            enabled_mgr_objs.append(mgr_obj)
    for res_obj in enabled_res_objs:
        res_obj.enable_auto_update()
        assert res_obj.auto_update_enabled()
    for mgr_obj in enabled_mgr_objs:
        mgr_obj.enable_auto_update()
        assert mgr_obj.auto_update_enabled()

    # pylint: disable=protected-access
    updater = faked_session._auto_updater
    if not auto_updated_uris:
        assert not updater.is_open()
        return

    assert updater.is_open()
    assert updater.has_objects()
    assert faked_session.auto_update_subscribed()

    for res_obj in enabled_res_objs:
        registered_objs = list(updater.registered_objects(res_obj.uri))
        assert res_obj in registered_objs

    # Simulate the desired changes in the mock enviroment
    for parent_uri, resources in added_resources:
        if parent_uri is None:
            faked_parent_obj = faked_session.hmc
        else:
            faked_parent_obj = faked_session.hmc.lookup_by_uri(parent_uri)
        faked_parent_obj.add_resources(resources)
    for parent_uri, mgr_attr, res_oid in removed_resources:
        if parent_uri is None:
            faked_parent_obj = faked_session.hmc
        else:
            faked_parent_obj = faked_session.hmc.lookup_by_uri(parent_uri)
        faked_mgr_obj = getattr(faked_parent_obj, mgr_attr)
        faked_mgr_obj.remove(res_oid)
    for res_uri, props in updated_resources:
        faked_res_obj = faked_session.hmc.lookup_by_uri(res_uri)
        faked_res_obj.properties.update(props)

    # Queue the notifications to be sent for the desired changes
    stomp_conn = updater._conn  # pylint: disable=protected-access
    for headers, message in notifications:
        # pylint: disable=no-member
        stomp_conn.mock_add_message(headers, message)

    # Send the notifications to the auto updater
    stomp_conn.mock_start()  # pylint: disable=no-member

    # Wait for notifications to be received and processed.
    time.sleep(0.5)

    # Verify the log entries.
    # The expected log entries must appear in the specified order, but
    # additional log entries are tolerated.
    caplog_records = list(caplog.records)
    next_aix = 0
    for eix, exp_log_entry in enumerate(exp_log_entries):
        exp_logger_name, exp_level, exp_message_pattern = exp_log_entry
        for aix, act_record in enumerate(caplog_records[next_aix:]):
            if act_record.name == exp_logger_name \
                    and act_record.levelno == exp_level \
                    and re.match(exp_message_pattern, act_record.message):
                # Found the next matching expected log entry
                next_aix += 1 + aix
                break
        else:
            exp_messages = [item[2] for item in exp_log_entries]
            caplog_messages = [item.message for item in caplog_records]
            exp_msg_str = '\n'.join(exp_messages)
            caplog_msg_str = '\n'.join(caplog_messages)
            raise AssertionError(
                f"Did not find expected log message pattern at index {eix} "
                f"in captured log records starting at index {next_aix}.\n"
                "\nExpected log record message patterns:\n"
                f"{exp_msg_str}\n"
                "\nCaptured log record messages:\n"
                f"{caplog_msg_str}")

    # Verify the list() results and the local resources in the manager object
    for mgr_uri, exp_res_uris, exp_local_res_uris in exp_lists:
        mgr_obj = managers[mgr_uri]

        # Verify the list() results
        res_objs = mgr_obj.list()
        res_uris = [_obj.uri for _obj in res_objs]
        assert set(res_uris) == set(exp_res_uris)

        # Verify the local resources in the manager object
        local_res_objs = mgr_obj.list_resources_local()
        local_res_uris = [_obj.uri for _obj in local_res_objs]
        assert set(local_res_uris) == set(exp_local_res_uris)

    # Verify the properties of updated resource objects
    for res_uri, exp_props in exp_resources:
        faked_res_obj = faked_session.hmc.lookup_by_uri(res_uri)
        for name, value in faked_res_obj.properties.items():
            try:
                exp_value = exp_props[name]
            except KeyError:
                exp_value = None
            if exp_value is not None:
                assert value == exp_value, (
                    f"Unexpected value for property {name} of resource "
                    f"{faked_res_obj.uri}")

    # Disable the resource and manager objects for auto-updating.
    # This causes them to be unregistered from the auto updater.
    for res_obj in enabled_res_objs:
        res_obj.disable_auto_update()
        assert not res_obj.auto_update_enabled()
    for mgr_obj in enabled_mgr_objs:
        mgr_obj.disable_auto_update()
        assert not mgr_obj.auto_update_enabled()

    # There is always the initial CPC manager object in the UpdateListener
    # class enabled for auto-updating:
    assert updater.has_objects()
    assert updater.is_open()
    assert faked_session.auto_update_subscribed()
